import os
import sys


class PyTestUtils:
    @staticmethod
    def is_pytest_execution():
        if "pytest" in sys.modules and "PYTEST_CURRENT_TEST" in os.environ:
            return True
        return False



import logging
import os

LOG = logging.getLogger(__name__)


class GitUtils:

    @classmethod
    def does_git_repo_dir_exist(cls, project):
        path = os.path.join(project.target_dir, ".git")
        if os.path.exists(path):
            return True
        return False

    @classmethod
    def convert_remote_branch_name_to_local(cls, remote_branch):
        stripped_rbranch = remote_branch.lstrip()
        # Strip off leading "<remote>/" part, if any
        split_parts = stripped_rbranch.rsplit("/", 1)

        if len(split_parts) == 2:
            l_branch = split_parts[1]
        else:
            # Branch is already in local format
            l_branch = split_parts

        return l_branch

    @classmethod
    def get_number_of_conflicts_from_str(cls, str):
        conflicts = str.count("patch failed: ")
        LOG.debug("Number of conflicts: %s", conflicts)
        return conflicts



import logging
from urllib.parse import urlparse

import requests

LOG = logging.getLogger(__name__)


class UrlUtils:

    @staticmethod
    def get_hostname_from_url(url):
        # from urlparse import urlparse  # Python 2
        parsed_uri = urlparse(url)
        return '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)

    @staticmethod
    def get_url_components(url):
        # from urlparse import urlparse  # Python 2
        parsed_uri = urlparse(url)
        # hostname = {str} 'szyszy.ddns.net'
        # netloc = {str} 'szyszy.ddns.net:33000'
        # params = {str} ''
        # password = {NoneType} None
        # path = {str} '/'
        # port = {int} 33000
        # query = {str} ''
        # scheme = {str} 'http'
        # username = {NoneType} None
        return {
            "hostname": parsed_uri.hostname,
            "netloc": parsed_uri.netloc,
            "scheme": parsed_uri.scheme,
        }

    @staticmethod
    def url_ok(url, silent=False):
        print_exc_info = not silent
        try:
            r = requests.head(url)
        except requests.exceptions.RequestException as e:

            LOG.exception("Failed to connect to URL: {}".format(url), exc_info=print_exc_info)
            return False
        return r.status_code == 200



import logging
import os
from enum import Enum
from typing import List

from git import Repo, RemoteProgress, GitCommandError, Commit, Actor
from pythoncommons.git_constants import ORIGIN
from pythoncommons.git_constants import HEAD, COMMIT_FIELD_SEPARATOR
from pythoncommons.git_utils import GitUtils
from pythoncommons.jira_wrapper import PatchOverallStatus, AdvancedJiraPatch, PatchApply, JiraPatchStatus

FORMAT_CODE_HASH = "%H"
FORMAT_CODE_COMMIT_MSG = "%s"
FORMAT_CODE_DATE_ISO_8601 = "%cI"
FORMAT_CODE_AUTHOR = "%ae"
FORMAT_CODE_COMMITTER = "%ce"
DEFAULT_BRANCH = "master"

LOG = logging.getLogger(__name__)


# TODO GitLogLineFormat could store the actual log format strings
class GitLogLineFormat(Enum):
    ONELINE_WITH_DATE = 0
    ONELINE_WITH_DATE_AND_AUTHOR = 1
    ONELINE_WITH_DATE_AUTHOR_COMMITTER = 2


class GitWrapper:
    def __init__(self, base_path):
        self.repo_path = base_path
        self.repo = Repo(self.repo_path)

    # TODO Similar functionality here: yarndevtools.yarn_dev_tools.Setup._setup_gitpython_log
    @property
    def is_enabled_git_cmd_logging(self):
        return self.repo.git.GIT_PYTHON_TRACE is not False

    def enable_debug_logging(self, full=False):
        value = "full" if full else "1"
        os.environ["GIT_PYTHON_TRACE"] = value
        # https://github.com/gitpython-developers/GitPython/issues/222#issuecomment-68597780
        type(self.repo.git).GIT_PYTHON_TRACE = value

    def get_current_branch_name(self):
        return self.repo.git.rev_parse("HEAD", symbolic_full_name=True, abbrev_ref=True)

    def checkout_branch(self, branch, track=False):
        prev_branch = self.get_current_branch_name()
        LOG.info("Checking out branch: %s (Previous branch was: %s)", branch, prev_branch)
        if branch not in self.repo.heads:
            raise ValueError(f"Cannot find branch: {branch}")

        self.repo.git.checkout(branch)
        if track:
            LOG.info("Tracking branch '%s' with remote '%s'", branch, ORIGIN)
            self.repo.git.branch("-u", ORIGIN + "/" + branch)

    def move_branch(self, branch, move_to="HEAD"):
        if move_to == "HEAD":
            self.repo.git.branch("-f", branch)
        else:
            self.repo.git.branch("-f", branch, move_to)

    def checkout_new_branch(self, new_branch, base_ref):
        base_exist = self.is_branch_exist(base_ref)
        if not base_exist:
            return False

        prev_branch = self.get_current_branch_name()
        LOG.info(
            "Checking out new branch: %s based on ref: %s (Previous branch was: %s)", new_branch, base_ref, prev_branch
        )
        try:
            self.repo.git.checkout(base_ref, b=new_branch)
        except GitCommandError:
            LOG.exception("Git checkout failed!", exc_info=True)
            return False
        return True

    def pull(self, remote_name, ff_only=False):
        progress = ProgressPrinter("pull")
        remote = self.repo.remote(name=remote_name)
        LOG.info("Pulling remote: %s", remote_name)

        kwargs = {}
        if ff_only:
            kwargs["ff-only"] = True
        try:
            result = remote.pull(progress=progress, **kwargs)
            LOG.debug("Result of git pull: %s", result)
        except GitCommandError:
            LOG.exception("Failed to execute git command. Printing some diagnostic info...", exc_info=True)
            branch = self.get_current_branch_name()
            LOG.error("Current branch: %s", branch)
            branch_tup = GitWrapper._get_branch_tuple(branch)
            git_branches_arg = "{}..{}".format(branch_tup[0], branch_tup[1])
            git_log_out = self.repo.git.execute(["git", "log", git_branches_arg])
            LOG.error("Git commit diff between branches: %s: %s", git_branches_arg, git_log_out)

    @staticmethod
    def _get_branch_tuple(branch):
        remote_br = ORIGIN + "/" + branch
        return branch, remote_br

    def checkout_and_pull(self, checkout_ref, remote_to_pull=ORIGIN, no_ff=False, ff_only=False):
        pull_kwargs = {}
        if no_ff:
            pull_kwargs["no-ff"] = True
        if ff_only:
            pull_kwargs["ff-only"] = True

        self.checkout_branch(checkout_ref)
        LOG.info(f"Pulling {remote_to_pull}")
        self.repo.remotes[remote_to_pull].pull(**pull_kwargs)

    def fetch(self, repo_url=None, remote_name=None, all=False):
        progress = ProgressPrinter("fetch")
        if not repo_url and not remote_name and not all:
            raise ValueError("Please specify remote or use the 'all' switch")

        if repo_url:
            try:
                LOG.info("Fetching from provided repo URL: %s", repo_url)
                self.repo.git.fetch(repo_url, remote_name)
                return True
            except GitCommandError:
                LOG.exception("Git fetch failed.", exc_info=True)
                return False

        if all:
            LOG.info("Fetching all remotes of git repo '%s'...", self.repo_path)
            for remote in self.repo.remotes:
                LOG.info("Fetching remote '%s' of repository: %s", remote, self.repo.git_dir)
                remote.fetch()
        else:
            LOG.info("Fetching remote '%s' of repository: %s", remote_name, self.repo.git_dir)
            remote = self.repo.remote(name=remote_name)
            remote.fetch(progress=progress)

    def checkout_previous_branch(self):
        prev_branch = self.get_current_branch_name()
        self.repo.git.checkout("-")
        LOG.info("Checked out: %s (Previous branch was: %s)", self.get_current_branch_name(), prev_branch)

    def rebase(self, rebase_onto):
        LOG.info("Rebasing onto branch: %s", rebase_onto)

        try:
            self.repo.git.rebase(rebase_onto)
        except GitCommandError:
            LOG.exception("Rebase failed!", exc_info=True)
            try:
                self.abort_rebase()
                LOG.error("Rebase was aborted! Please rebase manually!")
            except GitCommandError as e2:
                LOG.debug("Rebase was not in progress, but probably this is normal. Exception data: %s", e2)
            return False

        return True

    def abort_rebase(self):
        LOG.info("Aborting rebase...")
        self.repo.git.rebase(abort=True)

    def diff_check(self, raise_exception=True):
        try:
            self.repo.git.diff(check=True)
            return True
        except GitCommandError as e:
            LOG.error("Git diff --check failed. There are trailing whitespaces in the diff, please fix them!")
            if raise_exception:
                raise e
            return False

    def apply_check(self, patch, raise_exception=False):
        try:
            self.repo.git.apply(patch, check=True)
            return True
        except GitCommandError as e:
            LOG.exception("Git apply --check failed", exc_info=True)
            if raise_exception:
                raise e
            return False

    def apply_patch(self, patch, include_check=True, raise_exception=False):
        try:
            if include_check:
                self.apply_check(patch, raise_exception=False)
            LOG.info("Applying patch: %s", patch)
            self.repo.git.apply(patch)
            return True
        except GitCommandError as e:
            LOG.error("Git apply failed for patch %s!", patch)
            if raise_exception:
                raise e
            return False

    def apply_patch_advanced(self, patch: AdvancedJiraPatch, branch_prefix):
        if not isinstance(patch, AdvancedJiraPatch):
            raise ValueError("patch must be an instance of JiraPatch!")
        if not self.repo:
            raise ValueError("Repository is not yet synced! Please invoke sync_hadoop method before this method!")

        LOG.info("Applying patch %s on branches: %s", patch.filename, patch.target_branches)
        LOG.debug("Applying patch %s", patch)

        results = []
        for branch in patch.target_branches:
            patch_branch_name = "{prefix}-{branch}-{filename}".format(
                prefix=branch_prefix, branch=branch, filename=patch.filename
            )
            target_branch = "origin/" + branch

            if not patch.is_applicable_for_branch(branch):
                LOG.warning(
                    "Patch %s is not applicable on branch %s! Reason: %s!",
                    patch,
                    branch,
                    patch.get_reason_for_non_applicability(branch),
                )
                results.append(PatchApply(patch, target_branch, JiraPatchStatus.PATCH_ALREADY_COMMITTED))
                continue

            # If branch already exists, move it to target_branch
            if patch_branch_name in self.repo.heads:
                LOG.info(
                    "Patch branch already exists with name %s, moving branch pointer to %s",
                    patch_branch_name,
                    target_branch,
                )
                patch_branch = self.repo.heads[patch_branch_name]
                patch_branch.set_commit(target_branch)
            else:
                patch_branch = self.repo.create_head(patch_branch_name, target_branch)

            self.repo.head.reference = patch_branch
            self.cleanup()
            try:
                LOG.debug("[%s] Applying patch %s to branch: %s...", patch.issue_id, patch.filename, target_branch)
                status, stdout, stderr = self.repo.git.execute(
                    ["git", "apply", patch.file_path], with_extended_output=True
                )
                self.log_git_exec(status, stderr, stdout)
                if status == 0:
                    LOG.info(
                        "[%s] Successfully applied patch %s to branch: %s.",
                        patch.issue_id,
                        patch.filename,
                        target_branch,
                    )
                    results.append(PatchApply(patch, target_branch, JiraPatchStatus.APPLIES_CLEANLY))
                else:
                    LOG.error("Something bad happened")
                    self.log_git_exec(status, stderr, stdout, level=logging.INFO)
            except GitCommandError as gce:
                if "patch does not apply" in gce.stderr:
                    LOG.info("[%s] Patch %s does not apply to %s!" % (patch.issue_id, patch.filename, target_branch))
                    self.log_git_exec(gce.status, gce.stderr, gce.stdout)

                    conflicts = GitUtils.get_number_of_conflicts_from_str(gce.stderr)
                    results.append(
                        PatchApply(
                            patch,
                            target_branch,
                            JiraPatchStatus.CONFLICT,
                            conflicts=conflicts,
                            conflict_details=gce.stderr,
                        )
                    )
                else:
                    results.append(PatchApply(patch, target_branch, JiraPatchStatus.UNKNOWN_ERROR))

        return results

    def validate_branches(self, branches):
        if not self.repo:
            raise ValueError("Repository is not yet synced! Please invoke sync_hadoop method before this method!")
        for branch in branches:
            Repo.rev_parse(self.repo, "origin/" + branch)

    def cleanup(self):
        self.repo.head.reset(index=True, working_tree=True)
        self.repo.git.clean("-xdfq")

    def log_git_exec(self, status, stderr, stdout, level=logging.DEBUG):
        if level == logging.DEBUG:
            LOG.debug("Status of git command: %s", status)
            LOG.debug("stdout of git command: %s", stdout)
            LOG.debug("stderr of git command: %s", stderr)
        else:
            LOG.info("Status of git command: %s", status)
            LOG.info("stdout of git command: %s", stdout)
            LOG.info("stderr of git command: %s", stderr)

    def get_commit_hashes(self, issue_id, branch=None):
        cmd_list = ["git", "log"]
        basic_options = ["--oneline", "--grep", issue_id]
        if branch:
            cmd_list.append(branch)
        else:
            cmd_list.append("--all")
        cmd_list.extend(basic_options)

        status, stdout, stderr = self.repo.git.execute(cmd_list, with_extended_output=True)
        self.log_git_exec(status, stderr, stdout)
        if status != 0:
            raise ValueError("[%s] Failed to run git log command that finds a Jira issue!")
        if stdout:
            commit_hashes = []
            for line in stdout.splitlines():
                line_parts = line.split(" ")
                if len(line_parts) > 0:
                    commit_hashes.append(line_parts[0])
            return commit_hashes

        return []

    def get_remote_branches_for_commits(self, commits, strip_remote=True):
        if commits is None:
            raise ValueError("List of commits should not be None!")

        remote_branches = []
        for commit in commits:
            status, stdout, stderr = self.repo.git.execute(
                ["git", "branch", "-r", "--contains", commit], with_extended_output=True
            )
            self.log_git_exec(status, stderr, stdout)
            if status != 0:
                raise ValueError("[%s] Failed to run git branch command that finds remote branches for commit!")
            if stdout:
                for r_branch in stdout.splitlines():
                    if len(r_branch) > 0:
                        stripped_rbranch = r_branch.lstrip()

                        if strip_remote:
                            local_branch = GitUtils.convert_remote_branch_name_to_local(r_branch)
                            remote_branches.append(local_branch)
                        else:
                            remote_branches.append(stripped_rbranch)
            else:
                return []
        return remote_branches

    def diff(self, branch, cached=False):
        kwargs = {}
        if cached:
            kwargs["cached"] = True

        LOG.info("Making diff against %s", branch)
        return self.repo.git.diff(branch, **kwargs)

    def diff_between_refs(self, ref1, ref2):
        LOG.info("Making diff: %s..%s", ref1, ref2)
        return self.repo.git.diff(f"{ref1}..{ref2}")

    def diff_tree(self, ref, no_commit_id=None, name_only=None, recursive=False):
        args = [ref]

        kwargs = {}
        if no_commit_id:
            kwargs["no_commit_id"] = True
        if name_only:
            kwargs["name_only"] = True
        if recursive:
            kwargs["r"] = True

        if not self.is_enabled_git_cmd_logging:
            LOG.info("Running git diff-tree with arguments, args: %s, kwargs: %s", args, kwargs)
        diff_tree_results = self.repo.git.diff_tree(*args, **kwargs).splitlines()
        return diff_tree_results

    def show(self, hash, no_patch=None, no_notes=None, pretty=None, suppress_diff=False, format=None):
        args = [hash]

        kwargs = {}
        if no_patch:
            kwargs["no_patch"] = True
        if no_notes:
            kwargs["no_notes"] = True
        if pretty:
            kwargs["pretty"] = pretty
        if suppress_diff:
            kwargs["s"] = True
        if format:
            kwargs["format"] = format

        if not self.is_enabled_git_cmd_logging:
            LOG.info("Running git show with arguments, args: %s, kwargs: %s", args, kwargs)
        result = self.repo.git.show(*args, **kwargs).splitlines()
        return result

    def get_author_by_hash(self, hash):
        return self.show(hash, suppress_diff=True, format="%ae")

    def is_working_directory_clean(self):
        status = self.repo.git.status(porcelain=True)
        LOG.debug("Git status: %s", status)
        return False if len(status) > 0 else True

    def is_branch_exist(self, branch: str, exc_info=True):
        try:
            self.repo.git.rev_parse("--verify", branch)
            return True
        except GitCommandError:
            LOG.exception("Branch does not exist", exc_info=exc_info)
            return False

    def list_branches(self, name):
        try:
            branches = self.repo.git.branch("--list", name)
            branches = branches.split("\n")
            return [b.replace(" ", "") for b in branches]
        except GitCommandError:
            LOG.exception(f"Branch does not exist with name: {name}", exc_info=True)
            return []

    def add_all_and_commit(self, commit_msg, raise_exception=False):
        try:
            self.repo.git.add("-A")
            self.repo.index.commit(commit_msg)
            return True
        except GitCommandError:
            LOG.exception("Failed to commit changes from index", exc_info=True)
            return False

    def commit(
        self, message, author: Actor = None, committer: Actor = None, add_files_to_index: List[str] = None, amend=False
    ):
        if not add_files_to_index:
            add_files_to_index = []

        kwargs = {}
        if amend:
            kwargs["amend"] = amend
        if author:
            kwargs["author"] = author
        if committer:
            kwargs["committer"] = committer
        LOG.info("Running git commit with arguments: %s", kwargs)

        if add_files_to_index:
            self.add_to_index(add_files_to_index)
        self.repo.index.commit(message, **kwargs)

    def log(
        self,
        revision_range,
        oneline=False,
        oneline_with_date=False,
        oneline_with_date_and_author=False,
        oneline_with_date_author_committer=False,
        grep=None,
        format=None,
        n=None,
        return_hashes=False,
        return_messages=False,
        as_string_message=False,
        follow=False,
        all=False,
        grep_first_line_only=False,
    ):
        # TODO Raise error if any of oneline_with_date, oneline_with_date_and_author, oneline_with_date_author_committer
        # are True at the same time
        if oneline and oneline_with_date:
            raise ValueError("oneline and oneline_with_date should be exclusive!")

        if as_string_message and n != 1:
            raise ValueError("as_string_message only works with option n=1")

        args = []
        if revision_range:
            args.append(revision_range)

        kwargs = {}
        if oneline:
            kwargs["oneline"] = True

        # https://git-scm.com/docs/pretty-formats
        # Oneline format: <hash> <title line>
        # Oneline + date format: <hash> <title line> <author date>
        if oneline_with_date:
            kwargs["format"] = f"{FORMAT_CODE_HASH} {FORMAT_CODE_COMMIT_MSG} {FORMAT_CODE_DATE_ISO_8601}"
        if oneline_with_date_and_author:
            kwargs[
                "format"
            ] = f"{FORMAT_CODE_HASH} {FORMAT_CODE_COMMIT_MSG} {FORMAT_CODE_DATE_ISO_8601} {FORMAT_CODE_AUTHOR}"
        if oneline_with_date_author_committer:
            kwargs[
                "format"
            ] = f"{FORMAT_CODE_HASH} {FORMAT_CODE_COMMIT_MSG} {FORMAT_CODE_DATE_ISO_8601} {FORMAT_CODE_AUTHOR} {FORMAT_CODE_COMMITTER}"
        if format:
            kwargs["format"] = format
        if grep:
            kwargs["grep"] = grep
        if n:
            kwargs["n"] = n
        if follow:
            kwargs["follow"] = True
        if all:
            kwargs["all"] = True

        if not self.is_enabled_git_cmd_logging:
            LOG.info("Running git log with arguments, args: %s, kwargs: %s", args, kwargs)
        log_result = self.repo.git.log(*args, **kwargs).splitlines()

        if grep and grep_first_line_only and grep not in log_result:
            log_result = []

        if return_hashes:
            return [result.split(COMMIT_FIELD_SEPARATOR)[0] for result in log_result]
        if return_messages:
            # Remove commit hash and rejoin parts of commit message into one string
            # TODO Below command only works for --oneline option
            return [COMMIT_FIELD_SEPARATOR.join(result.split(COMMIT_FIELD_SEPARATOR)[1:]) for result in log_result]

        if as_string_message:
            return "\n".join(log_result)
        return log_result

    def branch(self, refspec, recursive=False, contains=None):
        args = []
        if refspec:
            args.append(refspec)

        kwargs = {}
        if recursive:
            kwargs["r"] = True
        if contains:
            kwargs["contains"] = contains

        lines = self.repo.git.branch(refspec, *args, **kwargs).splitlines()
        for idx, line in enumerate(lines):
            # Replace all whitespace with empty string
            lines[idx] = "".join(line.split())
        return lines

    def cherry_pick(self, ref, x=False):
        kwargs = {}
        if x:
            kwargs["x"] = x
        try:
            self.repo.git.cherry_pick(ref, **kwargs)
            return True
        except GitCommandError:
            LOG.exception("Failed to cherry-pick commit: " + ref, exc_info=True)
            return False

    def format_patch(self, revision_range, output_dir=None, full_index=None):
        # git format-patch ${GIT_BASE_BRANCH} --output-directory ${GIT_FORMAT_PATCH_OUTPUT_DIR} --full-index
        args = []
        if revision_range:
            args.append(revision_range)

        kwargs = {}
        if output_dir:
            kwargs["output_directory"] = output_dir
        if full_index:
            kwargs["full_index"] = True

        if not self.is_enabled_git_cmd_logging:
            LOG.info("Running git format-patch with arguments, args: %s, kwargs: %s", args, kwargs)
        return self.repo.git.format_patch(*args, **kwargs)

    def rewrite_head_commit_message(self, prefix=None, postfix=None):
        if not prefix and not postfix:
            raise ValueError("You must provide either prefix or postfix!")
        old_commit_msg = self.get_head_commit_message()
        # Add downstream (CDH jira) number as a prefix.
        # Since it triggers a commit, it will also add gerrit Change-Id to the commit.
        self.repo.git.commit(amend=True, message=f"{prefix}{old_commit_msg}")

    def merge_base(self, feature_br: str, master_br: str) -> List[Commit]:
        return self.repo.merge_base(feature_br, master_br)

    def get_head_commit_message(self):
        return self.log(HEAD, format="%B", n=1, as_string_message=True)

    def get_commit_message_of_branch(self, branch):
        # TODO error handling if branch does not exist
        commit = self.repo.heads[branch].commit
        actual_commit_message = commit.message.rstrip()
        return actual_commit_message

    @staticmethod
    def extract_commit_hash_from_gitlog_results(results):
        return [res.split(COMMIT_FIELD_SEPARATOR)[0] for res in results]

    @staticmethod
    def extract_commit_hash_from_gitlog_result(result):
        return result.split(COMMIT_FIELD_SEPARATOR)[0]

    def reset_changes(self, reset_to=DEFAULT_BRANCH, reset_index=True, reset_working_tree=True, clean=True):
        LOG.info(f"Reset all changes. Params: {locals()}")
        self.repo.head.reset(commit=reset_to, index=reset_index, working_tree=reset_working_tree)
        if clean:
            self.repo.git.clean("-xdf")

    def reset(self, hard=False):
        if hard:
            self.repo.git.reset("--hard")
        else:
            self.repo.git.reset()

    def setup_committer_info(self, user, email):
        self.repo.config_writer().set_value("user", "name", user).release()
        self.repo.config_writer().set_value("user", "email", email).release()

    def remove_committer_info(self):
        self.repo.config_writer().set_value("user", "name", "").release()
        self.repo.config_writer().set_value("user", "email", "").release()

    def setup_pull_mode_ff_only(self, global_mode=False):
        config_level = self._get_git_config_level(global_mode)
        self.repo.config_writer(config_level=config_level).set_value("pull", "ff", "only").release()

    def read_config(self, global_mode=False):
        config_level = self._get_git_config_level(global_mode)
        conf_reader = self.repo.config_reader(config_level)
        conf_reader.read()
        self._safe_get_config_section(conf_reader, "pull", config_level)
        self._safe_get_config_section(conf_reader, "push", config_level)

    @staticmethod
    def _safe_get_config_section(conf_reader, section_name, config_level):
        try:
            conf = conf_reader.items_all(section_name)
            LOG.info("Git config for section '%s': %s", section_name, conf)
            return conf
        except KeyError:
            LOG.warning("Section '%s' does not exist in Git config. Config mode: %s", section_name, config_level)

    @staticmethod
    def _get_git_config_level(global_mode: bool):
        config_level = "repository"
        if global_mode:
            config_level = "global"
        return config_level

    def add_remote(self, name, url):
        try:
            self.repo.create_remote(name, url=url)
        except GitCommandError:
            # TODO make swallowing exception optional
            pass

    def remove_remote(self, name):
        self.repo.delete_remote(name)

    def get_hash_of_commit(self, branch):
        return self.repo.heads[branch].commit.hexsha

    def checkout_parent_of_branch(self, branch):
        if branch not in self.repo.heads:
            raise ValueError(f"Cannot find branch: {branch}")
        self.repo.git.checkout(branch + "^")
        return self.repo.git.rev_parse("--verify", HEAD)

    def get_all_branch_names(self):
        return [br.name for br in self.repo.heads]

    def remove_branches_with_prefix(self, prefix, checkout_before_remove=DEFAULT_BRANCH):
        branches = self.get_all_branch_names()
        matching_branches = list(filter(lambda br: br.startswith(prefix), branches))

        for branch in matching_branches:
            self.remove_branch(branch, checkout_before_remove=checkout_before_remove)

    def remove_branch(self, branch, ignore_error=True, checkout_before_remove=DEFAULT_BRANCH):
        LOG.info("Removing branch: %s", branch)
        # Checkout default branch, in case of branch is currently checked out
        self.checkout_branch(checkout_before_remove)

        try:
            self.repo.delete_head(branch, force=True)
        except GitCommandError as e:
            if not ignore_error:
                raise e

    def add_to_index(self, items: List[str]):
        LOG.debug(f"Adding files to index: {items}")
        self.repo.index.add(items)

    def __str__(self):
        filtered_dict = dict(filter(lambda elem: elem[0] in ["repo_path"], vars(self).items()))
        return "%s(%s)" % (type(self).__name__, ", ".join("%s=%s" % item for item in filtered_dict.items()))


class ProgressPrinter(RemoteProgress):
    def __init__(self, operation):
        super(ProgressPrinter, self).__init__()
        self.operation = operation

    def update(self, op_code, cur_count, max_count=None, message=""):
        percentage = cur_count / (max_count or 100.0) * 100
        LOG.info("Progress of git %s: %s%% (speed: %s)", self.operation, percentage, message or "-")



import os
import shutil
import unittest
from pythoncommons.file_utils import FileUtils


class FileUtilsTests(unittest.TestCase):
    def setUp(self) -> None:
        self.link_container_dir = "/tmp/link_container"
        self.linked_dir = "/tmp/linked_dir"
        self.link_name = "link_name"
        if os.path.exists(self.link_container_dir):
            shutil.rmtree(self.link_container_dir)
        os.umask(0)

    def tearDown(self) -> None:
        pass

    def test_create_symlink_path_dir_link_dest_not_exist(self):
        FileUtils.create_symlink_path_dir(self.link_name, self.linked_dir, self.link_container_dir)

    def test_create_symlink_path_dir_link_dest_exist_but_link_does_not_point_to_file(self):
        os.makedirs(self.link_container_dir, mode=0o777)
        full_path_to_link = FileUtils.join_path(self.link_container_dir, self.link_name)
        FileUtils.create_new_empty_file(full_path_to_link)
        FileUtils.create_symlink_path_dir(self.link_name, self.linked_dir, self.link_container_dir)

    def test_create_symlink_path_dir_link_dest_exist_and_link_points_to_file(self):
        os.makedirs(self.link_container_dir, mode=0o777)
        full_path_to_link = FileUtils.join_path(self.link_container_dir, self.link_name)
        FileUtils.create_new_empty_file(full_path_to_link)

        # Remove link, otherwise os.symlink won't work with existing files
        FileUtils.remove_file(full_path_to_link)
        os.symlink(self.linked_dir, full_path_to_link)
        FileUtils.create_symlink_path_dir(self.link_name, self.linked_dir, self.link_container_dir)

    def test_read_file_into_list(self):
        text_file_path = "/tmp/pythontest/textfile"
        FileUtils.create_new_empty_file(text_file_path)
        FileUtils.write_to_file(text_file_path, "bla\nbla2\nbla3")
        file_lines_list = FileUtils.read_file_to_list(text_file_path)
        self.assertEqual(['bla', 'bla2', 'bla3'], file_lines_list)



import inspect
import logging
import pickle
from typing import Any, List, Dict, Tuple, Iterable

LOG = logging.getLogger(__name__)


class ObjUtils:
    @staticmethod
    def print_properties(obj):
        LOG.info("Printing properties of obj: %s", obj)
        for prop, value in vars(obj).items():
            LOG.info("%s: %s", prop, value)

    @staticmethod
    def get_class_members(clazz):
        attributes = inspect.getmembers(clazz, lambda a: not (inspect.isroutine(a)))
        return [a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))]

    @staticmethod
    def get_properties(obj):
        return [(prop, value) for prop, value in vars(obj).items()]

    @staticmethod
    def get_static_fields(clazz):
        return [v for v, m in vars(clazz).items() if not (v.startswith("_") or callable(m))]

    @staticmethod
    def get_static_fields_with_values(clazz):
        return {v: getattr(clazz, v) for v, m in vars(clazz).items() if not (v.startswith("_") or callable(m))}

    @staticmethod
    def ensure_all_attrs_present(obj, attrs: List[Tuple[str, str]]):
        not_found_attrs: List[Tuple[str, str]] = []

        for attr_tup in attrs:
            attr_name = attr_tup[0]
            if not hasattr(obj, attr_name):
                not_found_attrs.append(attr_tup)

        if not not_found_attrs:
            return
        if len(not_found_attrs) == 1:
            tup: Tuple[str, str] = not_found_attrs[0]
            raise ValueError("Attribute '{}' (name: {}) is not specified!".format(*tup))
        elif len(not_found_attrs) > 1:
            exc_message = "The following attributes are not specified: \n"
            for tup in not_found_attrs:
                exc_message += "Attribute '{}' (name: {})\n".format(*tup)
            raise ValueError(exc_message)

    @staticmethod
    def ensure_all_attrs_with_value(obj, attrs: Iterable[str], expected_value):
        not_found_attrs: List[str] = []
        attrs_with_different_value: List[Tuple[str, Any]] = []

        for attr_name in attrs:
            if not hasattr(obj, attr_name):
                not_found_attrs.append(attr_name)
            elif getattr(obj, attr_name) != expected_value:
                attrs_with_different_value.append((attr_name, getattr(obj, attr_name)))

        if not_found_attrs:
            raise ValueError("The following attributes are not found: {}".format(not_found_attrs))
        if attrs_with_different_value:
            raise ValueError(
                "The following attributes are not having value of '{}': {}".format(
                    expected_value, attrs_with_different_value
                )
            )


class ListUtils:
    @staticmethod
    def get_duplicates(lst: List[Any]):
        seen = set()
        dupes = set()
        for x in lst:
            if x not in seen:
                seen.add(x)
            elif x not in dupes:
                dupes.add(x)
        return dupes

    @staticmethod
    def split_to_chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]


class CollectionUtils:
    @staticmethod
    def sum_len_of_lists_in_dict(d: Dict):
        return sum([len(lst) for lst in d.values()])

    @staticmethod
    def partition(pred, iterable):
        trues = []
        falses = []
        for item in iterable:
            if pred(item):
                trues.append(item)
            else:
                falses.append(item)
        return trues, falses

    @staticmethod
    def partition_multi(predicates, iterable) -> List[List[Any]]:
        lists = []
        for i in range(len(predicates)):
            lists.append([])

        for item in iterable:
            for idx, pred in enumerate(predicates):
                if pred(item):
                    lists[idx].append(item)
        return lists

    @staticmethod
    def get_dict_keys_with_prefix(dictionary: Dict[Any, Any], prefix: str):
        return [key.startswith(prefix) for key in dictionary.keys()]

    @staticmethod
    def pairwise(iterable):
        "s -> (s0, s1), (s2, s3), (s4, s5), ..."
        a = iter(iterable)
        return zip(a, a)


class PickleUtils:
    @staticmethod
    def dump(data, file):
        with open(file, "wb") as f:
            pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(file):
        with open(file, "rb") as f:
            # The protocol version used is detected automatically, so we do not
            # have to specify it.
            return pickle.load(f)



import logging
import unittest

from pythoncommons.constants import PROJECT_NAME
from pythoncommons.file_utils import FileUtils
from pythoncommons.jira_utils import JiraUtils
from pythoncommons.project_utils import ProjectUtils

LOG = logging.getLogger(__name__)


class JiraUtilsTests(unittest.TestCase):
    def test_YARN_10496(self):
        jira_html_file, jira_id, jira_list_file = self.init_jira_files("YARN-10496")
        jira_html = JiraUtils.download_jira_html(
            "https://issues.apache.org/jira/browse/", jira_id, jira_html_file
        )
        jira_ids_and_titles = JiraUtils.parse_subjiras_and_jira_titles_from_umbrella_html(
            jira_html, jira_list_file, filter_ids=[jira_id]
        )

        expected_jira_ids = {'YARN-10169', 'YARN-10504', 'YARN-10506', 'YARN-10512', 'YARN-10513',
                             'YARN-10521', 'YARN-10522', 'YARN-10524', 'YARN-10525', 'YARN-10531', 'YARN-10532',
                             'YARN-10535', 'YARN-10564', 'YARN-10571', 'YARN-10573', 'YARN-10574',
                             'YARN-10576', 'YARN-10577', 'YARN-10578', 'YARN-10579', 'YARN-10581', 'YARN-10587', 'YARN-10596',
                             'YARN-10598', 'YARN-10599', 'YARN-10600', 'YARN-10604', 'YARN-10605', 'YARN-10609',
                             'YARN-10614', 'YARN-10615', 'YARN-10620', 'YARN-10622', 'YARN-10624'}
        actual_jira_ids = set(jira_ids_and_titles.keys())
        LOG.info("Expected jira ids: %s", expected_jira_ids)
        LOG.info("Actual jira ids: %s", actual_jira_ids)
        set_intersection = expected_jira_ids.intersection(actual_jira_ids)
        missing_jira_ids = expected_jira_ids.difference(actual_jira_ids)
        self.assertTrue(len(set_intersection) == len(expected_jira_ids),
                        "Difference of expected jira ids / actual jira ids sets should be empty! "
                        "Difference: " + str(missing_jira_ids))

        expected_mappings = {'YARN-10624': 'Support max queues limit configuration in new auto created queue, consistent with old auto created.'}
        self.assertEqual(expected_mappings['YARN-10624'], jira_ids_and_titles['YARN-10624'])
        self.assertTrue(isinstance(jira_ids_and_titles['YARN-10624'], str))

    @staticmethod
    def init_jira_files(jira_id: str):
        project_out_root = ProjectUtils.get_test_output_basedir(PROJECT_NAME, allow_python_commons_as_project=True)
        result_basedir = FileUtils.join_path(project_out_root, "jira-data")
        FileUtils.ensure_dir_created(result_basedir)
        jira_html_file = FileUtils.join_path(result_basedir, "jira.html")
        jira_list_file = FileUtils.join_path(result_basedir, "jira-list.txt")
        return jira_html_file, jira_id, jira_list_file

    def test_parse_umbrella_title(self):
        jira_html_file, jira_id, jira_list_file = self.init_jira_files("YARN-10496")
        jira_html = JiraUtils.download_jira_html(
            "https://issues.apache.org/jira/browse/", jira_id, jira_html_file
        )
        jira_title = JiraUtils.parse_jira_title(jira_html)
        self.assertEqual("[Umbrella] Support Flexible Auto Queue Creation in Capacity Scheduler", jira_title)



import json
import logging
import os
import re
from copy import copy
from dataclasses import field, dataclass
from enum import Enum
from typing import Dict, List, Pattern

from dataclasses_json import LetterCase, dataclass_json

from pythoncommons.date_utils import DateUtils
from pythoncommons.file_utils import JsonFileUtils
from pythoncommons.string_utils import auto_str

REGEX_DOT = "\\."

REGEX_TWO_DIGITS = "\\d\\d"
REGEX_FOUR_DIGITS = REGEX_TWO_DIGITS + REGEX_TWO_DIGITS

DEFAULT_PARSER_CONFIG_FILENAME = "parserconfig.json"
DEFAULT_PARSE_PREFIX_SEPARATOR = ":"
DEFAULT_ALLOWED_VALUES_SEPARATOR = ","
GREEDY_FIELD_POSTFIX = "_greedy"

LOG = logging.getLogger(__name__)

REGEX_FOR_STRING = "[a-zA-ZÀ-ú0-9-:@<>_@().,'|\\[\\]\\/]+"
REGEX_FOR_MULTI_WORD_STRING = '"[ a-zA-ZÀ-ú0-9-:@<>_@().,\'|\\[\\]\\/]+"'
REGEX_FOR_MULTI_WORD_STRING_GREEDY = "[ a-zA-ZÀ-ú0-9-:@<>_@().,'|\\[\\]\\/]+"


class FieldParseType(Enum):
    REGEX = "regex"
    REGEX_WITH_SMART_PARSE = "regexSmartParse"
    LITERAL = "literal"
    BOOL = "bool"
    INT = "int"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ExtractableField:
    parse_type: FieldParseType
    optional: bool
    value: str = None
    extract_inner_group: bool = False
    allowed_values: str or None = field(default=None)
    parse_prefix: str or None = field(default=None)
    precedence: int or None = field(default=100)
    eat_greedy_without_parse_prefix: bool or None = field(default=False)

    def __post_init__(self):
        if self.allowed_values:
            self.allowed_values_list = self.allowed_values.split(",")


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ParsedBlockFormat:
    format_string: str
    fields: Dict[str, ExtractableField] = field(default_factory=dict)
    variables: Dict[str, str] = field(default_factory=dict)
    mandatory_fields: List[str] = field(default_factory=list)
    FIELD_FORMAT: str = r"<([a-zA-Z0-9_ ]+)>"
    VAR_PATTERN: str = r"VAR\(([a-zA-Z_]+)\)"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GenericBlockParserSettings:
    parsed_block_format: ParsedBlockFormat
    date_formats: List[str] = field(
        default_factory=list
    )  # https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

    @property
    def fields_proxy(self):
        return self.parsed_block_format.fields

    @property
    def variables_proxy(self):
        return self.parsed_block_format.variables

    @property
    def mandatory_fields_proxy(self):
        return self.parsed_block_format.mandatory_fields

    @property
    def format_string_proxy(self):
        return self.parsed_block_format.format_string


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GenericBlockParserConfig:
    generic_parser_settings: GenericBlockParserSettings

    # dynamic properties
    field_positions: List[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GenericLineParserSettings:
    fields: Dict[str, ExtractableField] = field(default_factory=dict)
    variables: Dict[str, str] = field(default_factory=dict)
    date_formats: List[str] = field(
        default_factory=list
    )  # https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

    @property
    def fields_proxy(self):
        return self.fields

    @property
    def mandatory_fields_proxy(self):
        return []

    @property
    def variables_proxy(self):
        return self.variables

    @property
    def format_string_proxy(self):
        return None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GenericLineParserConfig:
    generic_parser_settings: GenericLineParserSettings


@auto_str
class ParserConfigReader:
    def __init__(self, data, obj_data_class, config_type):
        self.data = data
        self.obj_data_class = obj_data_class
        conf, ext_conf = self._parse(config_type)
        self.config: config_type = conf
        self.extended_config: obj_data_class = ext_conf
        self.generic_parser_settings = self.config.generic_parser_settings

        # Post init
        self._validate()
        self.config.date_regexes = self._convert_date_formats_to_patterns()

    def _validate(self):
        if not self.config.generic_parser_settings.date_formats:
            raise ValueError("No date format specified!")

        format_string = self.generic_parser_settings.format_string_proxy
        actual_field_names = frozenset((self.generic_parser_settings.fields_proxy.keys()))
        if format_string:
            self.config.field_positions = list(re.findall(ParsedBlockFormat.FIELD_FORMAT, format_string))
            expected_field_names = set(self.config.field_positions)

            if not expected_field_names or any([fn == "" for fn in expected_field_names]):
                raise ValueError(
                    "Expected field names is empty, this is not normal. Value: {}".format(expected_field_names)
                )

            diff = expected_field_names.difference(actual_field_names)
            if diff:
                raise ValueError(
                    "The following fields are not having the field config object {}. "
                    "Expected field names: {}".format(diff, expected_field_names)
                )
        else:
            self.config.field_positions = list(self.generic_parser_settings.fields_proxy.keys())

        diff = set(self.generic_parser_settings.mandatory_fields_proxy).difference(actual_field_names)
        if diff:
            raise ValueError(
                "Found unknown field names: {}. Allowed field names: {}".format(
                    diff, self.generic_parser_settings.mandatory_fields_proxy
                )
            )

        self._check_variables()
        self._validate_date_formats(self.config.generic_parser_settings.date_formats)
        self._validate_regexes()
        self._validate_alllowed_values()

    def _check_variables(self):
        for field_name, field_object in self.generic_parser_settings.fields_proxy.items():
            if field_object.parse_type != FieldParseType.REGEX:
                continue
            vars = re.findall(ParsedBlockFormat.VAR_PATTERN, field_object.value)
            vars_set = set(vars)
            if vars_set:
                LOG.debug("Find variables in field '%s': '%s'", field_name, field_object.value)
                available_vars = self.generic_parser_settings.variables_proxy
                diff = set(vars_set).difference(set(available_vars.keys()))
                if diff:
                    raise ValueError(
                        "Unknown variables '{}' in {}: {}. Available variables: {}".format(
                            diff, field_name, field_object.value, available_vars.keys()
                        )
                    )
                ParserConfigReader.resolve_variables(available_vars, field_name, field_object, vars_set)

    @staticmethod
    def resolve_variables(available_vars, field_name, field_object, vars_set):
        field_value = field_object.value
        LOG.debug("Resolving variables in string: %s", field_value)

        original_value = str(field_value)
        new_value = str(original_value)
        for var in vars_set:
            new_value = new_value.replace(f"VAR({var})", available_vars[var])
        LOG.debug("Resolved variables for '%s'. Old: %s, New: %s", field_name, original_value, new_value)
        field_object.value = new_value

    def _validate_regexes(self):
        for f in self.generic_parser_settings.fields_proxy.values():  # type: ExtractableField
            if f.extract_inner_group:
                if f.type != FieldParseType.REGEX:
                    raise ValueError(
                        "Invalid config. If 'extractInnerGroup' is enabled, field type should be regex. Field object: {}".format(
                            f
                        )
                    )
                elif not (self._ensure_there_is_only_one_regex_group(f.value)):
                    raise ValueError(
                        "Invalid config. If 'extractInnerGroup' is enabled, the regex should only have at most one group. Field object: {}".format(
                            f
                        )
                    )

    @staticmethod
    def _validate_date_formats(format_strings):
        for fmt in format_strings:
            LOG.debug("Formatting current date with format '%s': %s", fmt, DateUtils.now_formatted(fmt))
        # TODO

    def _validate_alllowed_values(self):
        for field_name, field_object in self.config.generic_parser_settings.fields_proxy.items():
            if field_object.parse_type == FieldParseType.BOOL and not field_object.allowed_values:
                raise ValueError(
                    "Parse type is set to '{}' on field '{}', but allowed values are not specified!".format(
                        FieldParseType.BOOL.value, field_name
                    )
                )

    @staticmethod
    def read_from_file(obj_data_class, config_type, dir=None, filename=None):
        if filename:
            parser_conf_file = filename
        elif dir:
            parser_conf_file = os.path.join(dir, DEFAULT_PARSER_CONFIG_FILENAME)
        else:
            parser_conf_file = DEFAULT_PARSER_CONFIG_FILENAME

        data_dict = JsonFileUtils.load_data_from_json_file(parser_conf_file)
        return ParserConfigReader(data_dict, obj_data_class, config_type)

    def _parse(self, config_type):
        generic_parser_config = config_type.from_json(json.dumps(self.data))
        LOG.info("Generic parser config: %s", generic_parser_config)

        extended_parser_config = self.obj_data_class.from_json(json.dumps(self.data))
        LOG.info("Extended parser config: %s", extended_parser_config)
        return generic_parser_config, extended_parser_config

    def _convert_date_formats_to_patterns(self):
        mappings = {"%m": REGEX_TWO_DIGITS, "%d": REGEX_TWO_DIGITS, "%Y": REGEX_FOUR_DIGITS, ".": REGEX_DOT}
        regexes: List[Pattern] = []
        for fmt in self.config.generic_parser_settings.date_formats:
            curr_regex = fmt
            for orig, pattern in mappings.items():
                curr_regex = curr_regex.replace(orig, pattern)
            curr_regex += "$"
            regexes.append(re.compile(curr_regex))
        return regexes

    @staticmethod
    def _ensure_there_is_only_one_regex_group(regex: str):
        count_open = regex.count("(")
        count_close = regex.count(")")
        if count_open != count_close:
            return False
        if count_open > 1 or count_open == 0:
            return False
        return True

    def __repr__(self):
        return self.__str__()


class RegexFieldMatchType(Enum):
    SINGLE_MATCH = "single"
    MATCH_ANYWHERE = "anywhere"


class RegexGenerator:
    MATCH_TYPE = RegexFieldMatchType.MATCH_ANYWHERE

    @staticmethod
    def get_regexes(field_objects: Dict[str, ExtractableField]):
        # Order dict by precedence
        field_objects = {k: v for k, v in sorted(field_objects.items(), key=lambda item: item[1].precedence)}

        regex_dict: Dict[str, List[str]] = {}
        additional_fields: Dict[str, ExtractableField] = {}
        used_group_names = {}
        for field_name, field_object in field_objects.items():
            group_name = field_name  # use uppercase field name everywhere
            if group_name not in used_group_names:
                used_group_names[group_name] = True
                regex_dict[group_name] = RegexGenerator._create_regexes(group_name, field_object)
            else:
                raise ValueError("Group name is already used in regex: {}".format(group_name))
            if field_object.eat_greedy_without_parse_prefix:
                field_key = group_name + GREEDY_FIELD_POSTFIX
                regex_dict[field_key] = RegexGenerator._create_regexes(
                    field_key, field_object, use_parse_prefix=False, greedy=True
                )
                copied_field = copy(field_object)
                additional_fields[field_key] = copied_field
        return regex_dict, additional_fields

    @staticmethod
    # TODO Only monthlyexpensesummarizer uses this, remove later
    def create_final_regex(parser_config):
        field_objects: Dict[str, ExtractableField] = parser_config.generic_parser_settings.fields_proxy
        final_regex = r""
        used_group_names = {}
        for field_name in parser_config.field_positions:
            field_object = field_objects[field_name]
            group_name = field_name
            if group_name not in used_group_names:
                used_group_names[group_name] = 1
                final_regex += RegexGenerator._create_regexes(group_name, field_object)
            else:
                if group_name not in parser_config.generic_parser_settings.mandatory_fields_proxy:
                    used_group_names[group_name] += 1
                    group_name = f"{group_name}_{used_group_names[group_name]}"
                    final_regex += RegexGenerator._create_regexes(group_name, field_object)
                else:
                    raise ValueError("Group name is already used in regex: {}".format(group_name))
        LOG.info("FINAL REGEX: %s", final_regex)
        return final_regex

    @staticmethod
    def _create_regexes(group_name, field_object: ExtractableField, use_parse_prefix=True, greedy=False) -> List[str]:
        regex_values: List[str] = [field_object.value]
        parse_prefix = ""
        if use_parse_prefix:
            parse_prefix = field_object.parse_prefix
            if not parse_prefix:
                parse_prefix = ""
            else:
                parse_prefix += DEFAULT_PARSE_PREFIX_SEPARATOR
        elif field_object.parse_type == FieldParseType.INT:
            regex_values = [r"\d+"]
        elif field_object.parse_type == FieldParseType.BOOL:
            regex_values = ["|".join(field_object.allowed_values_list)]

        if field_object.parse_type == FieldParseType.REGEX_WITH_SMART_PARSE:
            # Add another regex with quoted string + multiple word capability
            # example values could be parsed:
            # from: "word1 word2 word3"
            # from: word1
            if greedy:
                # NOTE: Order is important, need to match multi-word strings before single word string
                regex_values = [REGEX_FOR_MULTI_WORD_STRING_GREEDY, REGEX_FOR_STRING]
            else:
                regex_values = [REGEX_FOR_STRING, REGEX_FOR_MULTI_WORD_STRING]

        regex_values = RegexGenerator._create_prefixed_regexes(parse_prefix, regex_values)
        grouped_regexes = RegexGenerator._create_grouped_regexes(field_object, group_name, regex_values)
        return grouped_regexes

    @staticmethod
    def _create_prefixed_regexes(parse_prefix, regex_values):
        # TODO
        # if RegexGenerator.MATCH_TYPE == RegexFieldMatchType.MATCH_ANYWHERE:
        # return [f".*{parse_prefix}{r}" for r in regex_values]
        return [f"{parse_prefix}{r}" for r in regex_values]

    @staticmethod
    def _create_grouped_regexes(field_object, group_name, regex_values):
        if field_object.extract_inner_group:
            grouped_regexes = RegexGenerator._get_inner_group_grouped_regexes(group_name, regex_values)
            if field_object.optional:
                grouped_regexes = [f"({r})*" for r in grouped_regexes]
        else:
            grouped_regexes = [f"(?P<{group_name}>{r})" for r in regex_values]
            if field_object.optional:
                grouped_regexes = [f"{r}*" for r in grouped_regexes]
        return grouped_regexes

    @staticmethod
    def _get_inner_group_grouped_regexes(group_name, regex_values):
        ret = []
        for regex_value in regex_values:
            open_idx = regex_value.find("(")
            close_idx = regex_value.rfind(")")
            quantifier = regex_value[close_idx + 1]
            if quantifier not in ["*", "?", "+"]:
                quantifier = ""
            start = regex_value[:open_idx]
            end = regex_value[close_idx + 1 :]
            group = regex_value[open_idx + 1 : close_idx] + quantifier
            grouped_regex = f"(?P<{group_name}>{group})"
            new_regex_value = f"{start}{grouped_regex}{end}"
            ret.append(new_regex_value)
        return ret



import contextlib
import logging
import os
import site
import sys
import tempfile
import unittest

from pythoncommons.constants import REPO_ROOT_DIRNAME, PROJECT_NAME
from pythoncommons.file_utils import FileUtils, FindResultType
from pythoncommons.process import SubprocessCommandRunner
from pythoncommons.project_utils import SimpleProjectUtils
from pythoncommons.string_utils import StringUtils

TEST_PROJECT_NAME = "testproject"

LOG = logging.getLogger(__name__)
REPO_ROOT_DIR = FileUtils.find_repo_root_dir(__file__, REPO_ROOT_DIRNAME)


def get_test_scripts_dir():
    return SimpleProjectUtils.get_project_dir(
        basedir=REPO_ROOT_DIR, parent_dir=PROJECT_NAME, dir_to_find="test-scripts", find_result_type=FindResultType.DIRS
    )


TEST_SCRIPTS_DIR = get_test_scripts_dir()


class ProjectUtilsTests(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        _ = logging.getLogger(__name__)
        self.assertTrue(root_logger.handlers, msg="No handlers configured on root logger")

    def test_executing_script_from_uncommon_directory(self):
        LOG.debug("***TEST")
        with self._copy_script_to_temp_dir("hello_world_simple.py") as tup:
            tmp_dir = tup[0]
            script_abs_path = tup[1]
            sys.path.append(tmp_dir)
            proc = self.launch_script(script_abs_path)
            self.assertEqual(0, proc.returncode)

    def test_executing_script_from_uncommon_directory_additional_slashes_in_path(self):
        with self._copy_script_to_temp_dir("hello_world_simple.py", add_additional_slashes=True) as tup:
            tmp_dir = tup[0]
            script_abs_path = tup[1]
            sys.path.append(tmp_dir)
            proc = self.launch_script(script_abs_path)
            self.assertEqual(0, proc.returncode)

    def test_executing_script_from_global_user_site(self):
        script = self._copy_script_to_global_site(TEST_PROJECT_NAME, "test_project.py")
        # Second script (dummy_test_command.py) will be imported from the script
        self._copy_script_to_global_site(TEST_PROJECT_NAME, "test_project.py")
        self._copy_script_to_global_site(
            TEST_PROJECT_NAME, "dummy_test_command.py", relative_dest_dir=FileUtils.join_path("commands", "testcommand")
        )
        proc = self.launch_script(script)
        self.assertEqual(0, proc.returncode)

    @staticmethod
    def launch_script(script_abs_path):
        cmd = f"python3 {script_abs_path}"
        proc = SubprocessCommandRunner.run_and_follow_stdout_stderr(
            cmd, stdout_logger=LOG, exit_on_nonzero_exitcode=True
        )
        return proc

    @contextlib.contextmanager
    def _copy_script_to_temp_dir(self, script_filename: str, dest_filename: str = None, add_additional_slashes=True):
        if not dest_filename:
            dest_filename = script_filename

        # Can't use /tmp as it's platform dependent.
        # On MacOS, it's mounted as /private/tmp but some Linux systems don't have /private/tmp.
        # Let's use python's built-in temp dir creation methods.
        tmp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
        script_tmp_dir = FileUtils.join_path(tmp_dir.name, "python")
        FileUtils.ensure_dir_created(script_tmp_dir)
        script_abs_path = FileUtils.join_path(script_tmp_dir, dest_filename)

        if add_additional_slashes:
            split = script_abs_path.split(os.sep)
            size = len(split)
            split.insert(size - 1, os.sep)
            split.insert(size - 4, os.sep)
            script_abs_path = os.sep.join(split)
        src_script = FileUtils.join_path(TEST_SCRIPTS_DIR, script_filename)
        FileUtils.copy_file(src_script, script_abs_path)
        yield tmp_dir, script_abs_path

    @staticmethod
    def _copy_script_to_global_site(
        project_dir: str, script_filename: str, dest_filename: str = None, relative_dest_dir=None
    ):
        if not dest_filename:
            dest_filename = script_filename
        if not relative_dest_dir:
            relative_dest_dir = ""
        relative_dest_dirs = StringUtils.get_list_of_components_from_path(relative_dest_dir)

        python_global_site = site.getsitepackages()[0]
        script_dest_parent_dir = FileUtils.join_path(python_global_site, project_dir)
        FileUtils.ensure_dir_created(script_dest_parent_dir)
        LOG.info("Script parent dir: %s", script_dest_parent_dir)
        dest_file = FileUtils.join_path(script_dest_parent_dir, *relative_dest_dirs, script_filename)
        FileUtils.ensure_file_exists(dest_file, create=True)
        src_file = FileUtils.join_path(TEST_SCRIPTS_DIR, *relative_dest_dirs, dest_filename)
        FileUtils.copy_file(src_file, dest_file)
        return dest_file



import logging
import types
from enum import Enum
from typing import List, Sized, Callable

from pythoncommons.logging_setup import SimpleLoggingSetup

LOG = logging.getLogger(__name__)

COLLECTION_PLACEHOLDER = "$$coll$$"


class LoggingUtils:
    @staticmethod
    def print_logger_info(logger):
        if not logger:
            return
        print("Logger name: {}".format(logger.name))
        print("Logger effective level: {}".format(logging.getLevelName(logger.getEffectiveLevel())))
        print("Logger handlers: {}".format(logger.handlers))
        print("Logger disabled: {}".format(logger.disabled))
        print("Logger propagate: {}".format(logger.propagate))
        logger.info("here is a test record")
        LoggingUtils.print_logger_info(logger.parent)

    @staticmethod
    def ensure_loggers_are_on_level(loggers: List[str], level: int, project_name_prefix=None):
        for logger_name in loggers:
            logger = logging.getLogger(logger_name)
            LoggingUtils.ensure_logger_is_on_level(
                logger,
                level,
                raise_error_if_not_enabled_for=True,
                print_logger_info=True,
                print_additional_details=True,
                project_name_prefix=project_name_prefix,
            )

    @staticmethod
    def ensure_logger_is_on_level(
        logger: logging.Logger,
        level: int,
        raise_error_if_not_enabled_for=False,
        print_logger_info=True,
        print_additional_details=True,
        project_name_prefix=None,
    ):
        if not logger:
            return
        if print_logger_info:
            LoggingUtils.print_logger_info(logger)
        enabled = logger.isEnabledFor(level)
        if raise_error_if_not_enabled_for and not enabled:
            err_message = "Logger {} is not enabled for level {}. Current effective level is: {}".format(
                logger.name, logging.getLevelName(level), logging.getLevelName(logger.getEffectiveLevel())
            )

            if print_additional_details:
                logger_names, all_loggers = SimpleLoggingSetup.get_all_loggers_from_loggerdict()
                pythoncommons_loggers = SimpleLoggingSetup.get_pythoncommons_loggers(all_loggers)
                error_details = f"ALL LOGGERS: {all_loggers}\n" f"PYTHON COMMONS LOGGERS: {pythoncommons_loggers}"
                if project_name_prefix:
                    project_specific_loggers = SimpleLoggingSetup.get_project_specific_loggers(
                        all_loggers, project_name_prefix
                    )
                    error_details += f"\nPROJECT SPECIFIC LOGGERS: {project_specific_loggers}"
                err_message += f"\n{error_details}"
            raise ValueError(err_message)


class LoggerProperties(Enum):
    COMBINED_COLLECTION_LOGGING = (COLLECTION_PLACEHOLDER, "combined_log")

    def __init__(self, placeholder, func: str):
        self.placeholder = placeholder
        self.func = func


class LoggerFactory:
    @staticmethod
    def get_logger(name: str, props: List[LoggerProperties] = None):
        if not props:
            props = [LoggerProperties.COMBINED_COLLECTION_LOGGING]

        logger = logging.getLogger(name)
        for prop in props:
            real_func = getattr(LoggerFactory, prop.func)
            setattr(logger, prop.func, types.MethodType(real_func, logger))
        return logger

    def combined_log(
        self: logging.Logger,
        msg: str,
        coll: Sized = None,
        info_coll: Sized = None,
        debug_coll: Sized = None,
        info_coll_func: Callable[[Sized], str] = len,
        debug_coll_func: Callable[[Sized], str] = str,
        show_warnings: bool = False,
    ):
        if not any(e is not None for e in [coll, info_coll, debug_coll]):
            raise ValueError(
                "Wrong collection configuration, one of coll, info_coll or debug_coll should not be None!"
                "Please verify arguments!"
            )
        if coll is not None and not info_coll:
            info_coll = coll
        if coll is not None and not debug_coll:
            debug_coll = coll
        if not all([info_coll is not None, debug_coll is not None]):
            raise ValueError(
                "Wrong collection configuration, one of info_coll or debug_coll is None! " "Please verify arguments!"
            )

        level = self.getEffectiveLevel()
        # if self.level != logging.NOTSET and self.level not in [logging.INFO, logging.DEBUG]:
        if level not in [logging.INFO, logging.DEBUG]:
            LOG.error(
                f"Logger level was not among: {[logging.getLevelName(logging.INFO), logging.getLevelName(logging.DEBUG)]}"
            )
            return

        replace = True
        if COLLECTION_PLACEHOLDER not in msg:
            if show_warnings:
                LOG.warning(
                    f"Placeholder not found in message: {msg}. Will append collection to the end of the message."
                )
            replace = False

        len_info_coll = str(info_coll_func(info_coll))
        info_msg = f"{msg} {len_info_coll}"
        if replace:
            info_msg = msg.replace(COLLECTION_PLACEHOLDER, len_info_coll)

        if level == logging.INFO:
            self.info(info_msg)
        elif level == logging.DEBUG:
            debug_coll_str = debug_coll_func(debug_coll)
            if replace:
                debug_msg = msg.replace(COLLECTION_PLACEHOLDER, debug_coll_str)
            else:
                debug_msg = f"{msg} {debug_coll_str}"
            self.info(info_msg)
            self.debug(debug_msg)



import errno
import fnmatch
import hashlib
import logging
import os
import re
import shutil
import tempfile
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List, Dict, Any
from stat import ST_SIZE
from stat import ST_MTIME

import humanize

from pythoncommons.date_utils import timeit

LOG = logging.getLogger(__name__)


class FileMatchType(Enum):
    regex = "regex"
    fnmatch = "fnmatch"


class FindResultType(Enum):
    FILES = "files"
    DIRS = "dirs"


class FileFinderCriteria:
    def __init__(self, exclude_dirs, extension, regex_pattern, parent_dir, full_path_result):
        self.exclude_dirs = exclude_dirs
        self.extension = extension
        self.regex_pattern = regex_pattern
        self.full_path_result = full_path_result
        self.parent_dir = parent_dir


class FileFinder:
    old_debug: bool = False
    debug: bool = False
    LOG_PREFIX = "[FINDING FILES]"

    @classmethod
    def _smartlog(cls, s: str):
        if cls.debug:
            LOG.debug(f"{cls.LOG_PREFIX} {s}")

    @classmethod
    def _is_file_matches_criteria(cls, file, parent_dir, criteria: FileFinderCriteria):
        if (
            (criteria.extension and not file.endswith("." + criteria.extension))
            or (criteria.regex_pattern and not criteria.regex_pattern.match(file))
            or (criteria.parent_dir and not criteria.parent_dir == parent_dir)
        ):
            return False
        return True

    @staticmethod
    def _get_os_walk_kwargs(exclude_dirs):
        oswalk_kwargs: Dict[str, Any] = {}
        if exclude_dirs:
            # When topdown is True, the caller can modify the dirnames list in-place
            # (perhaps using del or slice assignment),
            # and walk() will only recurse into the subdirectories whose names remain in dirnames;
            # this can be used to prune the search
            oswalk_kwargs["topdown"] = True
        return oswalk_kwargs

    @classmethod
    def _handle_dir_exclusions(cls, dirs, exclude_dirs):
        if exclude_dirs:
            # Not enough to check against basename(root) as all other dirs underneath will be walked on the next
            # invocation of the walk generator with the loop statement
            orig_dirs = dirs.copy()
            dirs[:] = [d for d in dirs if d not in exclude_dirs]
            if len(orig_dirs) != len(dirs):
                cls._smartlog(f"Excluded dirs: {list(set(orig_dirs) - set(dirs))}")
        return dirs

    @classmethod
    def _find_files(cls, root, files, criteria: FileFinderCriteria) -> List[str]:
        result: List[str] = []
        for file in files:
            cls._smartlog(f"Processing file: {file}")
            if cls._is_file_matches_criteria(file, FileUtils.basename(root), criteria):
                cls._smartlog(f"File matched: {file}")
                if criteria.full_path_result:
                    result.append(FileUtils.join_path(root, file))
                else:
                    result.append(file)
        return result

    @classmethod
    def _find_dirs(cls, root, dirs, criteria: FileFinderCriteria) -> List[str]:
        result: List[str] = []
        for dir in dirs:
            cls._smartlog(f"Processing dir: {dir}")
            if cls._is_file_matches_criteria(dir, FileUtils.basename(root), criteria):
                cls._smartlog(f"Dir matched: {dir}")
                if criteria.full_path_result:
                    result.append(FileUtils.join_path(root, dir))
                else:
                    result.append(dir)
        return result

    @classmethod
    def find_files(
        cls,
        basedir: str,
        find_type: FindResultType,
        regex: str = None,
        parent_dir: str = None,
        single_level=False,
        full_path_result=False,
        extension=None,
        debug=False,
        exclude_dirs: List[str] = None,
        ensure_number_of_results: int = None,
    ):
        cls.old_debug = cls.debug
        cls.debug = debug
        find_criteria: FileFinderCriteria = cls._get_criteria_from_args(
            exclude_dirs, extension, regex, parent_dir, full_path_result
        )
        result_files: List[str] = []
        for root, dirs, files in os.walk(basedir, **FileFinder._get_os_walk_kwargs(exclude_dirs)):
            cls._smartlog(f"Processing root: {root}, dirs: {dirs}")
            dirs[:] = cls._handle_dir_exclusions(dirs, exclude_dirs)
            if find_type == FindResultType.FILES:
                result_files.extend(cls._find_files(root, files, find_criteria))
            elif find_type == FindResultType.DIRS:
                result_files.extend(cls._find_dirs(root, dirs, find_criteria))
            if single_level:
                return result_files
        cls.debug = cls.old_debug

        if ensure_number_of_results:
            if len(result_files) != ensure_number_of_results:
                raise ValueError(
                    "Number of results is not equal to expected result size! Expected: {}, Actual: {}".format(
                        ensure_number_of_results, len(result_files)
                    )
                )
        return result_files

    @classmethod
    def _get_criteria_from_args(cls, exclude_dirs, extension, regex, parent_dir, full_path_result):
        saved_args = locals().copy()
        cls._smartlog(f"Received args: {saved_args}")
        if not exclude_dirs:
            exclude_dirs = []
        # Preprocess
        if extension:
            if extension.startswith(".") or extension.startswith("*."):
                extension = extension.split(".")[-1]
            cls._smartlog(f"Filtering files with extension: {extension}")
        regex_pattern = re.compile(regex) if regex else None
        cls._smartlog(f"Modified args: {locals()}")
        return FileFinderCriteria(exclude_dirs, extension, regex_pattern, parent_dir, full_path_result)


class FileUtils:
    previous_cwd = None

    # TODO consolidate with save_to_file
    @classmethod
    def write_to_file(cls, file_path, data, bytes=False):
        file_access_mode = "w"
        if bytes:
            file_access_mode = "wb"
        f = open(file_path, file_access_mode)
        f.write(data)
        f.close()

    @classmethod
    def write_to_tempfile(cls, contents):
        tmp = tempfile.NamedTemporaryFile(delete=False)
        with open(tmp.name, "w") as f:
            f.write(contents)
        return tmp.name

    @classmethod
    def save_to_file(cls, file_path, contents):
        FileUtils.ensure_file_exists(file_path, create=True)
        file = open(file_path, "w")
        file.write(contents)
        file.close()

    @classmethod
    def append_to_file(cls, file_path, contents):
        file = open(file_path, "a")
        file.write(contents)
        file.close()

    @classmethod
    def append_data_to_file(cls, path, data):
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        file = open(path, "a")
        file.write(data)
        file.close()

    @classmethod
    def ensure_dir_created(cls, dirname, log_exception=False):
        """
        Ensure that a named directory exists; if it does not, attempt to create it.
        """
        try:
            os.makedirs(dirname)
        except OSError as e:
            if log_exception:
                LOG.exception("Failed to create dirs", exc_info=True)
            # If Errno is File exists, don't raise Exception
            if e.errno != errno.EEXIST:
                raise
        return dirname

    @classmethod
    def ensure_all_files_exist(cls, files):
        for file in files:
            if not os.path.isfile(file):
                raise Exception(file + " does not exist!")

    @classmethod
    def ensure_file_exists(cls, path, create=False):
        if not path:
            raise ValueError("Path parameter should not be None or empty!")

        if not create and not os.path.exists(path):
            raise ValueError("No such file or directory: {}".format(path))

        path_comps = path.split(os.sep)
        dirs = path_comps[:-1]
        dirpath = os.sep.join(dirs)
        if not os.path.exists(dirpath):
            LOG.info("Creating dirs: %s", dirpath)
            FileUtils.ensure_dir_created(dirpath, log_exception=False)

        if not os.path.exists(path):
            # Create empty file: https://stackoverflow.com/a/12654798/1106893
            LOG.info("Creating file: %s", path)
            open(path, "a").close()

    @classmethod
    def ensure_file_exists_and_readable(cls, file, verbose=False):
        if verbose:
            LOG.info("Trying to open file %s for reading..", file)
        f = open(file, "r")
        if not f.readable():
            raise ValueError("File {} is not readable".format(file))
        return file

    @classmethod
    def ensure_file_exists_and_writable(cls, file, verbose=False):
        if verbose:
            LOG.info("Trying to open file %s for writing..", file)
        f = open(file, "w")
        if not f.writable():
            raise ValueError("File {} is not readable".format(file))
        return file

    @classmethod
    def ensure_dir_is_writable(cls, d):
        is_dir = FileUtils.is_dir(d)
        if not is_dir:
            raise ValueError("%s is not a directory!", d)

        writable = os.access(d, os.W_OK)
        if not writable:
            raise ValueError("Directory is not writable: %s", d)

    @classmethod
    def ensure_parent_dir_is_writable(cls, f):
        path = Path(f)
        parent_dir = path.parent
        writable = os.access(parent_dir.__str__(), os.W_OK)
        if not writable:
            raise ValueError("Parent directory is not writable: %s", parent_dir)

    @classmethod
    def ensure_dir_is_empty(cls, d):
        is_dir = FileUtils.is_dir(d)
        if not is_dir:
            raise ValueError("%s is not a directory!", d)

        if not FileUtils.is_dir_empty(d):
            raise ValueError("Directory is not empty: %s", d)

    @classmethod
    def is_dir_empty(cls, d):
        children = os.listdir(d)
        if len(children) == 0:
            return True
        return False

    @staticmethod
    def search_files(basedir, filename):
        result = []
        for dp, dn, filenames in os.walk(basedir):
            for f in filenames:
                if f == filename:
                    result.append(os.path.join(dp, f))
        return result

    @staticmethod
    def search_dir(basedir, dirname):
        result = []
        for root, dirs, files in os.walk(basedir):
            for d in dirs:
                if d == dirname:
                    return os.path.join(root, d)
        return result

    @staticmethod
    def find_repo_root_dir(current_script: str, root_dir_name: str, raise_error=True):
        orig_path = os.path.realpath(current_script)
        path = orig_path
        visited = [path]
        while path != os.sep and not path.endswith(root_dir_name):
            path = FileUtils.get_parent_dir_name(path)
            visited.append(path)
        if path == os.sep:
            message = "Failed to find directory '{}' starting from path '{}'. " "Visited: {}".format(
                root_dir_name, orig_path, visited
            )
            if raise_error:
                raise ValueError(message)
            else:
                LOG.error(message)
        return path

    @staticmethod
    def find_repo_root_dir_auto(curr_file, files_to_search: List[str] = None, raise_error=True):
        def _does_files_exist_in_dir(d):
            if not FileUtils.is_dir(d, throw_ex=False):
                return False
            LOG.debug(f"Listing files in dir {d}")
            files = os.listdir(d)
            LOG.debug(f"Found files in dir '{d}': {files}")
            return all(f in files and FileUtils.is_file(FileUtils.join_path(d, f)) for f in files_to_search)

        if not files_to_search:
            files_to_search = ["pyproject.toml"]

        orig_path = os.path.realpath(curr_file)
        path = orig_path
        visited = [path]
        while path != os.sep and not _does_files_exist_in_dir(path):
            LOG.debug(f"Finding root dir: Current path is: {path}")
            path = FileUtils.get_parent_dir_name(path)
            LOG.debug(f"Finding root dir: Moving up the path, new path is: {path}")
            visited.append(path)

        if raise_error and path == os.sep:
            raise ValueError(
                f"Failed to find project root directory starting from path '{orig_path}'. " f"Visited: {visited}"
            )

        return path, visited

    # TODO rename method
    @classmethod
    def find_files(
        cls,
        basedir,
        find_type: FindResultType = FindResultType.FILES,
        regex: str = None,
        parent_dir: str = None,
        single_level=False,
        full_path_result=False,
        extension=None,
        debug=False,
        exclude_dirs: List[str] = None,
        ensure_number_of_results: int = None,
    ):
        args = locals().copy()
        del args["cls"]
        return FileFinder.find_files(**args)

    @staticmethod
    def list_files_in_dir(dir, pattern=None):
        LOG.info("Listing files in dir: " + dir)
        if not pattern:
            result = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        else:
            result = []
            for f in os.listdir(dir):
                file_path = os.path.join(dir, f)
                if os.path.isfile(file_path) and FileUtils.does_filename_match(f, pattern, FileMatchType.fnmatch):
                    result.append(file_path)
        return result

    @classmethod
    def does_filename_match(cls, filename, pattern, pattern_match_type):
        if pattern_match_type == FileMatchType.fnmatch and fnmatch.fnmatch(filename, pattern):
            return True
        elif pattern_match_type == FileMatchType.regex and re.search(pattern, filename, re.DOTALL):
            return True
        return False

    @classmethod
    def get_path_from_basedir(cls, basedir, path, include_last_dir=False):
        basedir_idx = path.rindex(basedir)
        start_idx = basedir_idx + len(basedir)
        if not basedir[-1] == os.sep:
            start_idx += len(os.sep)
        if include_last_dir:
            end_idx = len(path)
        else:
            end_idx = path.rindex(os.path.sep) + 1
        return path[start_idx:end_idx]

    @classmethod
    def remove_files(cls, dir, pattern):
        from pythoncommons.string_utils import RegexUtils

        if not FileUtils.does_file_exist(dir):
            LOG.warning("Directory does not exist: %s", dir)
            return
        for filename in os.listdir(dir):
            file_path = FileUtils.join_path(dir, filename)
            matches = RegexUtils.ensure_matches_pattern(FileUtils.path_basename(file_path), pattern)
            if not matches:
                LOG.debug("Filename not matched: %s", file_path)
                continue
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
                    LOG.debug("Successfully removed file: %s", file_path)
            except Exception as e:
                LOG.error("Failed to delete %s. Reason: %s", file_path, e)

    @classmethod
    def remove_file(cls, path):
        if not os.path.exists(path):
            LOG.warning("Cannot remove file as it does not exist: %s", path)
        os.remove(path)

    @classmethod
    def remove_all_files_in_dir(cls, dir, endswith="lock"):
        if not os.path.exists(dir):
            LOG.error("Can't delete files in dir as dir does not exist: %s", dir)
            return
        for filename in os.listdir(dir):
            file_path = os.path.join(dir, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    if endswith:
                        if file_path.endswith(endswith):
                            os.unlink(file_path)
                        else:
                            LOG.warning("Skip removing file, does not end with: %", endswith)
                    else:
                        os.unlink(file_path)
                # elif os.path.isdir(file_path):
                #     shutil.rmtree(file_path)
            except Exception as e:
                print("Failed to delete %s. Reason: %s" % (file_path, e))

    @classmethod
    def remove_dir(cls, dir, force=False):
        protected_dirs = ("/", "~")
        if dir in protected_dirs:
            LOG.warning(
                "Remove dir was invoked with directory: %s, which is in the protected dirs: %s", dir, protected_dirs
            )
            return
        if force:
            shutil.rmtree(dir, ignore_errors=True)
        else:
            os.rmdir(dir)

    @staticmethod
    def copy_files_to_dir(files: List[str], dst_dir: str, cut_path: str = None, cut_basedir: bool = False):
        LOG.debug("Copying files '%s' to destination dir: '%s'", files, dst_dir)
        for f in files:
            dest_filename = f
            if cut_path:
                if not f.startswith(cut_path):
                    raise ValueError(f"Expected cut_path '{cut_path}' to be in beginning of file's full path: {f}")
                dest_filename = FileUtils.join_path(*f.split(cut_path))
                if dest_filename.startswith(os.sep):
                    dest_filename = dest_filename[1:]
            if cut_basedir:
                dest_filename = os.path.basename(f)

            dest_file_path = os.path.join(dst_dir, dest_filename)
            FileUtils.ensure_file_exists(dest_file_path, create=True)
            LOG.debug("Copying %s to %s", f, dest_file_path)
            shutil.copyfile(f, dest_file_path)

    @staticmethod
    def copy_file_to_dir(src_file, dst_dir, dst_file_name_func, msg_template=None):
        dest_filename = dst_file_name_func(src_file, dst_dir)
        dest_file_path = os.path.join(dst_dir, dest_filename)

        if msg_template:
            LOG.info(msg_template.format(src_file, dest_file_path))
        shutil.copyfile(src_file, dest_file_path)
        return dest_file_path

    @classmethod
    def get_formatted_file_sizes_in_dir(cls, db_copies_dir, since: datetime = None):
        result = ""
        file_data = cls.get_file_sizes_with_mod_dates_in_dir(db_copies_dir)
        for fd in file_data:
            include = True
            if since:
                mod_date_of_file = datetime.fromtimestamp(float(fd[2]))
                if mod_date_of_file < since:
                    include = False
                    LOG.debug(f"Mod date of file < since, dropping it. File was: {fd[0]}")

            if include:
                human_readable_size = humanize.naturalsize(fd[1], gnu=True)
                result += "{size}    {file}\n".format(size=human_readable_size, file=fd[0])
        return result

    @classmethod
    def get_file_sizes_in_dir(cls, db_copies_dir):
        return cls._get_files_with_attrs_in_dir(db_copies_dir, [ST_SIZE])

    @classmethod
    def _get_files_with_attrs_in_dir(cls, db_copies_dir: str, stat_attrs_idx: List[int]):
        files = os.listdir(db_copies_dir)
        result = []
        for f in files:
            file_path = os.path.join(db_copies_dir, f)
            tup = (file_path,)
            for attr_idx in stat_attrs_idx:
                file_stats = os.stat(file_path)
                tup = tup + (file_stats[attr_idx],)
            result.append(tup)
        return result

    @classmethod
    def get_file_sizes_with_mod_dates_in_dir(cls, dir):
        return cls._get_files_with_attrs_in_dir(dir, [ST_SIZE, ST_MTIME])

    @classmethod
    def get_file_extension(cls, filename):
        filename, ext = os.path.splitext(filename)
        ext = ext.replace(".", "")
        return ext

    @classmethod
    def does_file_exist(cls, file):
        return os.path.exists(file)

    @classmethod
    def does_path_exist(cls, path):
        return os.path.exists(path)

    @classmethod
    def create_files(cls, *files):
        for file in files:
            FileUtils.ensure_file_exists(file, create=True)

    @classmethod
    def create_new_empty_file(cls, path):
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        file = open(path, "w")
        file.close()

    @classmethod
    def create_new_dir(cls, path, fail_if_created=True):
        if not os.path.exists(path):
            os.makedirs(path)
        elif fail_if_created:
            raise ValueError("Directory already exist: %s", path)

    @classmethod
    def verify_if_dir_is_created(cls, path, raise_ex=True):
        if not os.path.exists(path) or not os.path.isdir(path):
            if raise_ex:
                raise ValueError("Directory is not created under path: " + path)
            return False
        return True

    @classmethod
    def get_file_size(cls, file_path, human_readable=True):
        from pathlib import Path

        size = Path(file_path).stat().st_size
        if human_readable:
            return humanize.naturalsize(size, gnu=True)
        else:
            return str(size)

    @classmethod
    def check_size_match(cls, size, filepath):
        size_on_disk = FileUtils._get_file_size(filepath)
        return size_on_disk == int(size)

    @classmethod
    def check_size_less(cls, size, filepath):
        size_on_disk = FileUtils._get_file_size(filepath)
        return int(size) < size_on_disk

    @classmethod
    def _get_file_size(cls, f):
        return os.stat(f).st_size

    @classmethod
    def get_file_last_modified_date(cls, f):
        return os.path.getmtime(f)

    @classmethod
    def get_mod_date_of_file(cls, file):
        return os.path.getmtime(file)

    @classmethod
    def path_basename(cls, path):
        return os.path.basename(path)

    @classmethod
    def basename(cls, path):
        return os.path.basename(path)

    @classmethod
    def join_path(cls, *components):
        if components and components[0] and not components[0].startswith(os.sep) and not components[0].startswith("~"):
            lst = list(components)
            lst[0] = os.sep + components[0]
            components = tuple(lst)
        return os.path.join(*components)

    @classmethod
    def make_path(cls, basedir, dirs):
        if not isinstance(dirs, list):
            LOG.warning(
                "%s was called with wrong argument type for 'dirs', "
                "list is expected. Value of parameter: %s. "
                "Converting parameter to a list!",
                FileUtils.make_path,
                dirs,
            )
            dirs = [dirs]
        return os.path.join(basedir, *dirs)

    @classmethod
    def get_parent_dir_name(cls, dir):
        path = Path(dir)
        return path.parent.__str__()

    @classmethod
    def is_dir_parent_of_dir(cls, parent, dir):
        parent_path_parts = Path(parent).parts
        dir_path_parts = Path(dir).parts
        try:
            idx = dir_path_parts.index(parent_path_parts[0])
        except ValueError:
            # Do not log anything
            return False

        for i in range(len(dir_path_parts)):
            if len(parent_path_parts) - 1 >= i and not dir_path_parts[i] == parent_path_parts[idx + i]:
                return False
        return True

    @classmethod
    def get_path_components(cls, path):
        return path.rsplit(os.sep)

    @classmethod
    def get_mod_dates_of_files(cls, basedir, *files):
        result = {}
        for f in files:
            f = FileUtils.join_path(basedir, f)
            if FileUtils.does_file_exist(f):
                result[f] = FileUtils.get_mod_date_of_file(f)
            else:
                result[f] = None
        return result

    @classmethod
    def get_home_path(cls, path):
        return os.path.expanduser(path)

    @classmethod
    def copy_file(cls, src, dest):
        LOG.info(f"Copying file. {src} -> {dest}")
        shutil.copyfile(src, dest)

    @classmethod
    def _move_files(cls, src, dst):
        FileUtils.ensure_dir_created(dst)
        files = os.listdir(src)
        for f in files:
            src_path = os.path.join(src, f)
            dst_path = os.path.join(dst, f)
            LOG.info("Moving: {} --> {}", src_path, dst_path)
            os.rename(src_path, dst_path)

    @classmethod
    def change_cwd(cls, dir):
        cls.previous_cwd = os.getcwd()
        cls._change_cwd(dir)

    @classmethod
    def get_filename_from_cwd(cls, filename):
        return os.path.join(os.getcwd(), filename)

    @classmethod
    def reset_cwd(cls):
        if not cls.previous_cwd:
            LOG.warning("Can't reset CWD as there's no previous CWD saved!")
        cls._change_cwd(cls.previous_cwd)

    @classmethod
    def _change_cwd(cls, dir):
        try:
            os.chdir(dir)
            LOG.info("Changed current working directory: %s", dir)
        except OSError:
            LOG.error("Can't change the Current Working Directory to %s", dir)

    @classmethod
    def _hash_files_in_dirs(cls, dirs):
        hash_data = {}
        for dir in dirs:
            hash_data[dir] = {}
            files = os.listdir(dir)
            for f in files:
                file_to_hash = os.path.join(dir, f)
                hash = cls.hash_file(file_to_hash)
                hash_data[dir][hash] = f
        return hash_data

    @classmethod
    def hash_file(cls, f):
        blocksize = 65536
        with open(f, "rb") as file:
            hasher = hashlib.md5()
            buf = file.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = file.read(blocksize)
        return hasher.hexdigest()

    @classmethod
    def is_file(cls, f):
        if not os.path.exists(f):
            raise ValueError("Path does not exist: %s", f)
        return os.path.isfile(f)

    @classmethod
    def ensure_is_file(cls, f):
        is_file = FileUtils.is_file(f)
        if not is_file:
            raise ValueError("%s is not a file!", f)

    @classmethod
    def is_dir(cls, d, throw_ex=True):
        if not os.path.exists(d):
            if throw_ex:
                raise ValueError("Path does not exist: %s", d)
        return os.path.isdir(d)

    @classmethod
    def get_unique_filepath(cls, dest_file):
        while FileUtils.does_path_exist(dest_file):
            file_path, ext = os.path.splitext(dest_file)
            dest_file = file_path + "_1" + ext
        return dest_file

    @classmethod
    def read_file(cls, f):
        return open(f, "r").read()

    @classmethod
    def read_file_to_list(cls, f):
        return open(f, "r").read().splitlines()

    @classmethod
    def does_file_contain_str(cls, file, string):
        with open(file) as f:
            if string in f.read():
                return True
            return False

    @classmethod
    def create_symlink_path_dir(
        cls, link_name, linked_path, dest_dir, remove_link_if_exists=True, remove_linked_file_if_exists=False
    ):
        link_src = linked_path
        link_dest = FileUtils.join_path(dest_dir, link_name)
        if remove_link_if_exists:
            if os.path.islink(link_dest):
                LOG.info(f"Removing link dest: {link_dest}")
                os.unlink(link_dest)
            else:
                LOG.warning(f"Not removing not existing link dest: {link_dest}")
        if remove_linked_file_if_exists:
            if os.path.exists(link_src) and FileUtils.is_file(link_src):
                LOG.info(f"Removing linked file: {link_src}")
                FileUtils.remove_file(link_src)
            elif os.path.exists(link_src) and FileUtils.is_dir(link_src):
                LOG.info(f"Removing linked dir: {link_src}")
                shutil.rmtree(link_src)
            else:
                LOG.warning(f"Not removing not existing linked file or directory: {link_src}")

        FileUtils.create_symlink(link_src, link_dest)

    @classmethod
    def create_symlink(cls, link_src, link_dest):
        LOG.info("Creating symlink: %s -> %s", link_dest, link_src)
        # os.symlink(src, dest)
        # src: Already existing path to create the link pointing to
        # dest: Link name
        try:
            os.symlink(link_src, link_dest)
        except OSError as e:
            if e.errno == errno.EEXIST:
                LOG.warning("Symlink does exist, ignoring. Details: %s", str(e))

    @classmethod
    def get_temp_file_name(cls, prefix=None):
        kwargs = {"delete": False}
        if prefix:
            kwargs["prefix"] = prefix
        tmp = tempfile.NamedTemporaryFile(**kwargs)
        return tmp.name


class JsonFileUtils:
    @classmethod
    @timeit
    def write_data_to_file_as_json(cls, path, data, pretty=False):
        import json

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        if not os.path.isdir(dirname):
            raise ValueError("Should have a dir in path, not a file: {}".format(dirname))

        LOG.info("Starting to write to file: %s", path)
        with open(path, "w") as file:
            kwargs = {"sort_keys": True}
            if pretty:
                kwargs["indent"] = 4
            json.dump(data, file, **kwargs)
        LOG.info("Finished writing to file: %s", path)

    @classmethod
    def load_data_from_json_file(
        cls, file, create_if_not_exists=False, swallow_file_not_found=False, swallow_value_error=False
    ):
        import json

        try:
            with open(file, "r") as f:
                return json.load(f)
        except FileNotFoundError as e:
            LOG.exception("Error while opening file: %s", file)
            if create_if_not_exists:
                LOG.info("Creating new empty file: %s", file)
                FileUtils.create_new_empty_file(file)
            if swallow_file_not_found:
                return None
            raise e
        except ValueError as e:
            LOG.exception("Error while reading file: %s", file)
            if swallow_value_error:
                return None
            raise e


class CsvFileUtils:
    @classmethod
    def append_row_to_csv_file(cls, path, data, header=None):
        parent_dir = FileUtils.get_parent_dir_name(path)
        if not FileUtils.is_dir(parent_dir, throw_ex=False):
            FileUtils.create_new_dir(parent_dir)

        if not isinstance(data, list):
            raise ValueError("Expected list of data for CSV row!")

        new_file = True if not os.path.exists(path) else False

        with open(path, "a", newline="") as csvfile:
            import csv

            csv_writer = csv.writer(csvfile, delimiter=";", quotechar="|", quoting=csv.QUOTE_MINIMAL)
            if new_file and header:
                csv_writer.writerow(header)
            csv_writer.writerow(data)



import pprint
import re
from enum import Enum
import logging
from re import Pattern
from typing import Dict, Tuple, List, Any, Callable

from pythoncommons.file_parser.parser_config_reader import (
    RegexGenerator,
    DEFAULT_PARSE_PREFIX_SEPARATOR,
    GREEDY_FIELD_POSTFIX,
)
from pythoncommons.file_utils import FileUtils

LOG = logging.getLogger(__name__)


class DiagnosticInfoType(Enum):
    PARSED_OBJECTS = ("PARSED_OBJECT", "Parsed object: %s")
    MATCH_OBJECT = ("MATCH_OBJECT", "Match object: %s")
    MULTI_LINE_BLOCK_HEADER = ("MULTI_LINE_BLOCK_HEADER", "Found multi-line block header: %s")
    EXCLUDED_LINE = ("EXCLUDED_LINE", "Found excluded line: %s")
    MULTI_LINE_BLOCK = ("MULTI_LINE_BLOCK", "Found multi-line block: %s")
    SINGLE_LINE_BLOCK = ("SINGLE_LINE_BLOCK", "Found single-line block: %s")

    def __init__(self, value, log_pattern):
        self.log_pattern = log_pattern


class DiagnosticConfig:
    def __init__(
        self,
        print_date_lines: bool = False,
        print_multi_line_block_headers: bool = False,
        print_multi_line_blocks: bool = False,
        print_match_objs: bool = False,
        print_parsed_objects: bool = True,
        print_single_line_blocks: bool = True,
    ):
        self.print_match_objs = print_match_objs
        self.print_parsed_objects = print_parsed_objects
        self.print_date_lines = print_date_lines
        self.print_multi_line_block_headers = print_multi_line_block_headers
        self.print_multi_line_blocks = print_multi_line_blocks
        self.print_single_line_blocks = print_single_line_blocks
        self.conf_dict: Dict[DiagnosticInfoType, bool] = {
            DiagnosticInfoType.MULTI_LINE_BLOCK_HEADER: self.print_multi_line_block_headers,
            DiagnosticInfoType.EXCLUDED_LINE: self.print_date_lines,
            DiagnosticInfoType.MULTI_LINE_BLOCK: self.print_multi_line_blocks,
            DiagnosticInfoType.MATCH_OBJECT: self.print_match_objs,
            DiagnosticInfoType.PARSED_OBJECTS: self.print_parsed_objects,
            DiagnosticInfoType.SINGLE_LINE_BLOCK: self.print_single_line_blocks,
        }


class DiagnosticPrinter:
    def __init__(self, diagnostic_config: DiagnosticConfig):
        self.diagnostic_config = diagnostic_config

    def print_line(self, line, info_type: DiagnosticInfoType):
        enabled = self.diagnostic_config.conf_dict[info_type]
        if enabled:
            LOG.debug(info_type.log_pattern, line)

    def pretty_print(self, obj, info_type: DiagnosticInfoType):
        enabled = self.diagnostic_config.conf_dict[info_type]
        if enabled:
            LOG.debug(info_type.log_pattern, pprint.pformat(obj))


class GenericLineByLineParser:
    def __init__(self, generic_parser_config, diagnostic_config):
        self.generic_parser_config = generic_parser_config
        self.printer = DiagnosticPrinter(diagnostic_config)
        self._field_objects = self.generic_parser_config.generic_parser_settings.fields
        self.fields_by_regexes, additional_fields = RegexGenerator.get_regexes(self._field_objects)
        self._field_objects.update(additional_fields)
        LOG.info("Fields by regexes: %s", self.fields_by_regexes)
        self.lines_of_file = None

    def parse(self, file, parsed_object_dataclass: Any, line_to_obj_parser_func: Callable):
        file_contents = FileUtils.read_file(file)
        # TODO change to debug level
        LOG.info("File contents: %s\n", file_contents)
        self.lines_of_file = file_contents.split("\n")
        parsed_objects: List[parsed_object_dataclass] = []
        for idx, line in enumerate(self.lines_of_file):
            parsed_object = self._process_line(line, line_to_obj_parser_func)
            parsed_objects.append(parsed_object)
        self.printer.pretty_print(parsed_objects, DiagnosticInfoType.PARSED_OBJECTS)
        return parsed_objects

    def _process_line(self, line, line_to_obj_parser_func):
        matches: Dict[str, str] = {}
        for field_name, regexes in self.fields_by_regexes.items():
            for regex in regexes:
                LOG.debug("Trying to match field with name '%s' on line '%s' with regex '%s'", field_name, line, regex)
                match = re.search(regex, line)
                if match:
                    field_object = self._field_objects[field_name]
                    matched_str = match.group(field_name)
                    if not matched_str:
                        continue

                    result_str = matched_str
                    if field_object.parse_prefix:
                        prefix_with_sep = field_object.parse_prefix + DEFAULT_PARSE_PREFIX_SEPARATOR
                        if matched_str.startswith(prefix_with_sep):
                            result_str = matched_str[len(prefix_with_sep) :]
                            LOG.debug(
                                "Stripping prefix '%s' from string '%s', resulted string: '%s'",
                                prefix_with_sep,
                                matched_str,
                                result_str,
                            )
                    LOG.debug("Match: %s", result_str)
                    self._add_to_matches(field_name, field_object, matches, result_str)
                    self.printer.print_line(match, DiagnosticInfoType.MATCH_OBJECT)
                    line = line.replace(matched_str, "")
                    line = line.lstrip()
                    line = line.rstrip()
                else:
                    LOG.debug("Field with name '%s' on line '%s' with regex '%s' not found!", field_name, line, regex)
        LOG.debug("Final matches: %s", matches)
        return line_to_obj_parser_func(matches)

    @staticmethod
    def _add_to_matches(field_name, field_object, matches, result_str):
        if field_object.eat_greedy_without_parse_prefix:
            original_field_name = field_name.replace(GREEDY_FIELD_POSTFIX, "")
            if original_field_name not in matches or not matches[original_field_name]:
                matches[field_name] = result_str
        else:
            matches[field_name] = result_str


class GenericBlockBasedInputFileParser:
    def __init__(
        self, block_regex, block_open_chars, block_close_chars, diagnostic_config, excluded_line_patterns: List[Pattern]
    ):
        self.block_regex = block_regex
        self.printer = DiagnosticPrinter(diagnostic_config)
        self.block_open_chars = block_open_chars
        self.block_close_chars = block_close_chars
        self.block_definer = BlockDefiner(
            self.block_open_chars, self.block_close_chars, excluded_line_patterns, self.printer
        )
        self.line_ranges_of_blocks: List[Tuple[int, int]] = []
        self.excluded_line_patterns = excluded_line_patterns

    def parse(self, file: str, parsed_object_dataclass: Any, block_to_obj_parser_func: Callable):
        file_contents = FileUtils.read_file(file)
        self.lines_of_file = file_contents.split("\n")
        self.block_definer.define_blocks(self.lines_of_file)
        parsed_objects = self._process_line_ranges(parsed_object_dataclass, block_to_obj_parser_func)
        self.printer.pretty_print(parsed_objects, DiagnosticInfoType.PARSED_OBJECTS)
        return parsed_objects

    def _process_line_ranges(self, parsed_object_dataclass, block_to_obj_parser_func: Callable):
        self.lines_by_ranges: List[Tuple[List[str], str]] = self._get_lines_by_ranges()

        parsed_objects: List[parsed_object_dataclass] = []
        for list_of_lines, date in self.lines_by_ranges:
            lines = "\n".join(list_of_lines)
            match = re.match(self.block_regex, lines, re.MULTILINE)
            if not match:
                LOG.error("Block not matched: %s", lines)
                continue
            self.printer.print_line(match, DiagnosticInfoType.MATCH_OBJECT)
            parsed_objects.append(block_to_obj_parser_func(match, date))
        return parsed_objects

    def _get_lines_by_ranges(self):
        result: List[Tuple[List[str], str]] = []
        curr_date_idx = 0
        for range in self.block_definer.line_ranges_of_blocks:
            list_of_lines = self.lines_of_file[range[0] : range[1] + 1]
            if (len(self.block_definer.excluded_lines) - 1) != curr_date_idx and range[
                1
            ] > self.block_definer.excluded_lines[curr_date_idx + 1]:
                curr_date_idx += 1
            date_idx = self.block_definer.excluded_lines[curr_date_idx]
            date = self.lines_of_file[date_idx]
            result.append((list_of_lines, date))
        return result


class BlockDefiner:
    MULTI_LINE_BLOCK_CONTINUED = (-1, -1)
    MULTI_LINE_BLOCK_HEADER = (-2, -2)
    EMPTY_LINE = (-3, -3)

    def __init__(
        self, block_open_chars, block_close_chars, excluded_line_patterns: List[Pattern], printer: DiagnosticPrinter
    ):
        self.block_open_chars = block_open_chars
        self.block_close_chars = block_close_chars
        self.multiline_start_idx = -1
        self.multiline_end_idx = -1
        self.inside_multiline_block = False
        self.excluded_lines = []
        self.excluded_line_patterns = excluded_line_patterns
        self.printer = printer
        self.line_ranges_of_blocks = []

    def define_blocks(self, lines_of_file: List[str]):
        for idx, line in enumerate(lines_of_file):
            if self._determine_if_line_excluded(line):
                self.excluded_lines.append(idx)
            else:
                line_range = self._get_line_ranges_of_blocks(line, idx)
                if line_range not in (
                    BlockDefiner.EMPTY_LINE,
                    BlockDefiner.MULTI_LINE_BLOCK_CONTINUED,
                    BlockDefiner.MULTI_LINE_BLOCK_HEADER,
                ):
                    self.line_ranges_of_blocks.append(line_range)

    def _determine_if_line_excluded(self, line) -> bool:
        for excl_pattern in self.excluded_line_patterns:  # type: Pattern
            match = excl_pattern.match(line)
            if match:
                self.printer.print_line(line, DiagnosticInfoType.EXCLUDED_LINE)
                return True
        return False

    def _get_line_ranges_of_blocks(self, line, idx: int) -> Tuple[int, int]:
        if not line:
            LOG.debug("Encountered empty line")
            return BlockDefiner.EMPTY_LINE
        multi_line_opened: bool = any([char in line for char in self.block_open_chars])
        multi_line_closed: bool = any([char in line for char in self.block_close_chars])
        if multi_line_opened and not self.inside_multiline_block:
            self.inside_multiline_block = True
            self.multiline_start_idx = idx
            self.printer.print_line(line, DiagnosticInfoType.MULTI_LINE_BLOCK_HEADER)
            return self.MULTI_LINE_BLOCK_HEADER
        elif multi_line_closed and self.inside_multiline_block:
            self.inside_multiline_block = False
            self.multiline_end_idx = idx
            line_range = (self.multiline_start_idx, self.multiline_end_idx)
            self.printer.print_line(line_range, DiagnosticInfoType.MULTI_LINE_BLOCK)
            return line_range
        elif not multi_line_closed and self.inside_multiline_block:
            return self.MULTI_LINE_BLOCK_CONTINUED
        elif idx not in self.excluded_lines and (line and not line.isspace()):
            # Single line block
            line_range = (idx, idx)
            self.printer.print_line(line_range, DiagnosticInfoType.SINGLE_LINE_BLOCK)
            return line_range
        raise ValueError("Unexpected line: " + line)



import logging
import os
import re
import time
from enum import Enum
from typing import List, Tuple, Dict

import docker
from docker import APIClient
import json
from pythoncommons.file_utils import FileUtils
from pythoncommons.process import SubprocessCommandRunner
from pythoncommons.string_utils import auto_str

DEFAULT_DOCKERFILE_NAME = "Dockerfile"

LOG = logging.getLogger(__name__)


class DockerWrapper:
    client = APIClient(base_url="unix://var/run/docker.sock")

    def __init__(self):
        pass

    @classmethod
    def create_image_from_dir(cls, dockerfile_parent_dir_path, tag=None, build_args=None):
        cls._build_image_internal(dockerfile_parent_dir_path, tag=tag, build_args=build_args)

    @classmethod
    def create_image_from_dockerfile(cls, dockerfile_name, tag=None, build_args=None):
        dockerfile_parent_dir_path = os.path.dirname(dockerfile_name)
        dockerfile_name = os.path.basename(dockerfile_name)
        cls._build_image_internal(
            dockerfile_parent_dir_path, dockerfile_name=dockerfile_name, tag=tag, build_args=build_args
        )

    @classmethod
    def _build_image_internal(
        cls, dockerfile_parent_dir_path, dockerfile_name=DEFAULT_DOCKERFILE_NAME, tag=None, build_args=None
    ):
        if not build_args:
            build_args = {}
        LOG.info(
            "Starting to build Docker image from Dockerfile: %s, based on parent dir path: %s",
            dockerfile_name,
            dockerfile_parent_dir_path,
        )
        cls._fix_path_for_macos()
        response = [
            line
            for line in cls.client.build(
                path=dockerfile_parent_dir_path,
                dockerfile=dockerfile_name,
                rm=True,
                tag=tag,
                buildargs=build_args,
                network_mode="host",
            )
        ]
        errors = cls.log_response(response)
        if errors:
            raise ValueError(
                f"Failed to build Docker image from Dockerfile: {dockerfile_name}. " f"Error messages: {errors}"
            )

    @classmethod
    def _fix_path_for_macos(cls):
        # NOTE: To avoid docker.credentials.errors.InitializationError: docker-credential-osxkeychain
        # not installed or not available in PATH.
        # --> Must add /usr/local/bin/ to PATH on macosx platform
        current_path = os.environ["PATH"]
        if "/usr/local/bin" not in current_path:
            os.environ["PATH"] = current_path + ":/usr/local/bin"

    @classmethod
    def run_container(cls, image, volumes, sleep=300):
        client = docker.client.from_env()
        container = client.containers.run(image, "sleep {}".format(sleep), stderr=True, detach=True, volumes=volumes)
        return container

    @classmethod
    def inspect_container(cls, container_id: str):
        return docker.client.inspect_container(container_id)

    @classmethod
    def log_response(cls, response):
        errors = []
        for r in response:
            lines = r.decode().split("\r\n")
            for line in lines:
                if line:
                    line_dict = json.loads(line)
                    log_value = line_dict["stream"] if "stream" in line_dict else None
                    err_detail = line_dict["errorDetail"] if "errorDetail" in line_dict else None
                    if err_detail:
                        err_message = err_detail["message"]
                        errors.append(err_message)
                        LOG.error("[BUILD] %s", err_message)
                    if log_value and "ERROR" in log_value:
                        errors.append(log_value)
                    if log_value and (log_value != "\n"):
                        LOG.info("[BUILD] %s", log_value)
        return errors


class DockerMountMode(Enum):
    READ_WRITE = "rw"
    READ_ONLY = "ro"


@auto_str
class DockerMount:
    def __init__(self, host_dir, container_dir, mode=DockerMountMode.READ_WRITE):
        self.host_dir = host_dir
        self.container_dir = container_dir
        self.mode = mode


class DockerDiagnosticStdoutAssertionMode(Enum):
    EXACT_MATCH = "exact_match"
    SUBSTRING_MATCH = "substring_match"


class DockerDiagnosticPhase(Enum):
    PRE = "pre"
    POST = "post"


@auto_str
class DockerDiagnosticCommand:
    def __init__(
        self,
        mode,
        phase,
        command,
        expected_exit_code=0,
        expected_output=None,
        expected_output_fragments=None,
        strip=False,
    ):
        self.phase = phase
        self.mode = mode
        self.command = command
        self.expected_exit_code = expected_exit_code
        self.expected_output = expected_output
        self.expected_output_fragments = expected_output_fragments
        self.strip = strip

    @classmethod
    def create_exact_match(cls, phase, command, expected_output, expected_exit_code=0, strip=False):
        return cls(
            DockerDiagnosticStdoutAssertionMode.EXACT_MATCH,
            phase,
            command,
            expected_exit_code=expected_exit_code,
            expected_output=expected_output,
            strip=strip,
        )

    @classmethod
    def create_substring_match(cls, phase, command, expected_output_fragments, expected_exit_code=0, strip=False):
        return cls(
            DockerDiagnosticStdoutAssertionMode.SUBSTRING_MATCH,
            phase,
            command,
            expected_exit_code=expected_exit_code,
            expected_output_fragments=expected_output_fragments,
            strip=strip,
        )


class CreatePathMode(Enum):
    FULL_PATH = "FULL_PATH"
    PARENT_PATH = "PARENT_PATH"


class DockerTestSetup:
    def __init__(self, image_name, create_image=False, dockerfile_parent_dir_path=None, dockerfile=None, logger=None):
        self.image_name = image_name
        if create_image:
            if dockerfile_parent_dir_path:
                self.create_image(dockerfile_parent_dir_path=dockerfile_parent_dir_path)
            elif dockerfile:
                DockerWrapper.create_image_from_dockerfile(dockerfile, tag=self.image_name)

        if logger:
            self.CMD_LOG = logger
        else:
            self.CMD_LOG = LOG

        # Assigned later
        self._reinit()

    def _reinit(self):
        self.pre_diagnostics: List[DockerDiagnosticCommand] = []
        self.post_diagnostics: List[DockerDiagnosticCommand] = []
        self.test_instance = None
        self.mounts: List[DockerMount] = []
        self.container = None

    def cleanup(self):
        self._reinit()

    def create_image(self, dockerfile_parent_dir_path=None):
        if not dockerfile_parent_dir_path:
            dockerfile_parent_dir_path = os.getcwd()
            LOG.warning(
                f"Dockerfile location was not specified. "
                f"Trying to create image from current working directory: {dockerfile_parent_dir_path}"
            )
        DockerWrapper.create_image_from_dir(dockerfile_parent_dir_path, tag=self.image_name)

    def mount_dir(self, host_dir, container_dir, mode=DockerMountMode.READ_WRITE):
        self.mounts.append(DockerMount(host_dir, container_dir, mode=mode))

    def apply_mounts(self, docker_mounts: List[DockerMount]):
        self.mounts.extend(docker_mounts)

    def print_mounts(self):
        LOG.info("Docker mounts: %s", self.mounts)

    def add_diagnostics(self, diags: List[DockerDiagnosticCommand]):
        for diag in diags:
            if diag.phase == DockerDiagnosticPhase.PRE:
                self.pre_diagnostics.append(diag)
            elif diag.phase == DockerDiagnosticPhase.POST:
                self.post_diagnostics.append(diag)

    def run_container(self, commands_to_run: List[str] = None, sleep=300):
        if not commands_to_run:
            commands_to_run = []

        volumes_dict = self._create_volumes_dict()
        LOG.info(f"Starting container from image '{self.image_name}' with volumes: '{volumes_dict}'")
        self.container = DockerWrapper.run_container(image=self.image_name, volumes=volumes_dict, sleep=sleep)

        if self.pre_diagnostics:
            self._run_pre_diagnostic_commands()

        for cmd in commands_to_run:
            self.exec_cmd_in_container(cmd)

        if self.post_diagnostics:
            self._run_post_diagnostic_commands()
        return self.container

    def _create_volumes_dict(self):
        # Convert DockerMount objects to volumes dictionary
        volumes_dict = {}
        for mount in self.mounts:
            volumes_dict[mount.host_dir] = {"bind": mount.container_dir, "mode": mount.mode.value}
        return volumes_dict

    def _run_pre_diagnostic_commands(self):
        self._run_diagnostic_command(DockerDiagnosticPhase.PRE)

    def _run_post_diagnostic_commands(self):
        self._run_diagnostic_command(DockerDiagnosticPhase.POST)

    def _run_diagnostic_command(self, phase):
        diag_command_objs: List[DockerDiagnosticCommand] = (
            self.pre_diagnostics if phase == DockerDiagnosticPhase.PRE else self.post_diagnostics
        )
        LOG.debug("Running diagnostic commands in '%s' phase: %s", phase.value, diag_command_objs)
        for diag in diag_command_objs:
            if diag.mode == DockerDiagnosticStdoutAssertionMode.EXACT_MATCH:
                self.exec_diagnostic_command(diag)
            elif diag.mode == DockerDiagnosticStdoutAssertionMode.SUBSTRING_MATCH:
                self.exec_command_and_grep_in_stdout(diag)

    def exec_diagnostic_command(self, diag: DockerDiagnosticCommand):
        # TODO Seems like stdout is not returned anymore :(
        exit_code, stdout = self.exec_cmd_in_container(diag.command, strip=diag.strip)
        self.test_instance.assertEqual(
            diag.expected_exit_code,
            exit_code,
            msg="Exit code of command is not the expected. " f"Command details: {diag}",
        )
        if diag.strip:
            diag.expected_output = diag.expected_output.strip()
        self.test_instance.assertEqual(
            diag.expected_output, stdout, msg="Stdout of command is not the expected. " f"Command details: {diag}"
        )

    def exec_command_and_grep_in_stdout(self, diag: DockerDiagnosticCommand):
        # TODO Seems like stdout is not returned anymore :(
        exit_code, stdout = self.exec_cmd_in_container(diag.command, strip=diag.strip)
        self.test_instance.assertEqual(
            diag.expected_exit_code,
            exit_code,
            msg="Exit code of command is not the expected." f"Command details: {diag}",
        )

        for fragment in diag.expected_output_fragments:
            self.test_instance.assertTrue(
                fragment in stdout,
                msg="Cannot find expected fragment in stdout. "
                f"Fragment: {fragment}, stdout: {stdout}, Command details: '{diag}'",
            )

    def generate_dummy_text_files_in_container_dirs(self, dir_and_no_of_files: List[Tuple[str, int]]):
        for dir_files in dir_and_no_of_files:
            self._generate_dummy_text_files_in_container_dir(dir_files[0], dir_files[1])

    def _generate_dummy_text_files_in_container_dir(self, dir_path: str, number_of_files: int):
        self.exec_cmd_in_container("mkdir -p " + dir_path)
        for i in range(number_of_files):
            path = os.path.normpath(dir_path)
            path_segments = path.split(os.sep)
            path_segments = list(filter(None, path_segments))
            file_name = "_".join(path_segments) + "_" + str(i + 1)
            file_path = FileUtils.join_path(dir_path, file_name)
            cmd = f"echo dummy_{str(i + 1)} > {file_path}"
            # Simple redirect did not work: self._exec_cmd_in_container(cmd)
            # See: https://github.com/docker/docker-py/issues/1637
            # Use this as a workaround
            self.exec_cmd_in_container(["sh", "-c", cmd])

    def exec_cmd_in_container(
        self,
        cmd,
        charset="utf-8",
        strip=True,
        fail_on_error=True,
        stdin=False,
        tty=False,
        env: Dict[str, str] = None,
        detach=False,
        callback=None,
        stream=False,
        strict: bool = True,
    ):
        if not env:
            env = {}
        if strict:
            if not stream and callback:
                raise ValueError(
                    "Callback is specified but streaming mode is not enabled! Callback only makes sense if streaming mode is active!"
                )

        # https://stackoverflow.com/questions/29663459/python-app-does-not-print-anything-when-running-detached-in-docker
        env["PYTHONUNBUFFERED"] = "1"
        LOG.info(f"Running command '{cmd}' in container: '{self.container}'")
        exec_handler = DockerWrapper.client.exec_create(self.container.id, cmd, environment=env, stdin=stdin, tty=tty)
        ret = DockerWrapper.client.exec_start(exec_handler, stream=stream, detach=detach)

        # If stream=True, the execution will stay in _get_output_of_cmd until there's data to read from the output.
        # This means that when the loop that reads the output ends, the process is finished.
        # Therefore, handle stream mode separately.
        if not stream:
            self._get_and_verify_exit_code(callback, charset, cmd, exec_handler, fail_on_error, ret, stream, strip)

        if detach:
            exit_code: int = self._get_exit_code(cmd, exec_handler, stream)
            return exit_code, None

        decoded_stdout = self._get_output_of_cmd(cmd, ret, callback, charset, strip, stream)
        exit_code = self._get_and_verify_exit_code(
            callback, charset, cmd, exec_handler, fail_on_error, ret, stream, strip
        )
        return exit_code, decoded_stdout

    def _get_and_verify_exit_code(self, callback, charset, cmd, exec_handler, fail_on_error, ret, stream, strip):
        exit_code: int = self._get_exit_code(cmd, exec_handler, stream)
        if fail_on_error and exit_code != 0:
            _ = self._get_output_of_cmd(cmd, ret, callback, charset, strip, stream)
            raise ValueError(
                f"Command '{cmd}' returned with non-zero exit code: {exit_code}. See logs above for more details."
            )
        return exit_code

    def _get_output_of_cmd(self, cmd, ret, callback, charset, strip, stream):
        LOG.info(f"Listing stdout of cmd: {cmd}...")
        short_cmd = os.path.basename(cmd).rstrip()
        decoded_stdout = None

        if not stream:
            if ret:
                decoded_stdout = ret.decode(charset)
                if strip:
                    decoded_stdout = decoded_stdout.strip()
                self.CMD_LOG.info(f"[{short_cmd}] {decoded_stdout}")
                return decoded_stdout
            else:
                LOG.warning("Output was None")
                return None

        for output in ret:
            try:
                decoded_stdout = output.decode(charset)
                if strip:
                    decoded_stdout = decoded_stdout.strip()
                self.CMD_LOG.info(f"[{short_cmd}] {decoded_stdout}")
                if callback:
                    callback(cmd, decoded_stdout, self)
            except UnicodeDecodeError:
                LOG.error(f"Error while decoding string: {output.decode('cp437')}")
        return decoded_stdout

    @staticmethod
    def _get_exit_code(cmd: str, exec_handler, max_wait_seconds: int = 5):
        """
        client.exec_inspect(exec_handler["Id"]).get("ExitCode") does not immediately returns the exit code.
        Try to wait for it for some time.
        :param exec_handler:
        :return:
        """
        slept_seconds = 0
        while True:
            exit_code: int = DockerWrapper.client.exec_inspect(exec_handler["Id"]).get("ExitCode")
            LOG.debug("Command: '%s', exit code: %s", cmd, exit_code)
            if exit_code is not None:
                return exit_code
            else:
                LOG.debug("Commnand: '%s', exit code is still None. Sleeping 1s...")
                time.sleep(1)
                slept_seconds += 1
            if slept_seconds == max_wait_seconds:
                return None

    def inspect_container(self, container_id: str):
        return DockerWrapper.inspect_container(container_id)

    def docker_cp_from_container(self, container_path, local_target_path):
        command = f"docker cp {self.container.id}:{container_path} {local_target_path}"
        LOG.info(
            "Copying container directory '%s' to local directory '%s' (container id: %s). Command was: %s",
            container_path,
            local_target_path,
            self.container.id,
            command,
        )
        SubprocessCommandRunner.run_and_follow_stdout_stderr(command)

    def docker_cp_to_container(
        self,
        container_target_path,
        local_src_file,
        create_container_path_mode: CreatePathMode = None,
        double_check_with_ls: bool = False,
    ):
        # run mkdir -p if dir not exist
        self.create_directories_in_container(container_target_path, create_container_path_mode)
        command = f"docker cp {local_src_file} {self.container.id}:{container_target_path}"
        LOG.info(
            "Copying local directory '%s' to container directory '%s' (container id: %s). Command was: %s",
            local_src_file,
            container_target_path,
            self.container.id,
            command,
        )
        SubprocessCommandRunner.run_and_follow_stdout_stderr(command)
        if double_check_with_ls:
            self.exec_cmd_in_container(f"ls -la {container_target_path}")

    def create_directories_in_container(self, container_target_path: str, create_container_path_mode: CreatePathMode):
        if not create_container_path_mode:
            LOG.warning("Will not create directories as create_container_path_mode=%s", create_container_path_mode)
            return
        if create_container_path_mode:
            if create_container_path_mode == CreatePathMode.PARENT_PATH:
                path_to_create = FileUtils.basename(container_target_path)
                exit_code, _ = self.exec_cmd_in_container(f"ls {path_to_create}", fail_on_error=False)
            elif create_container_path_mode == CreatePathMode.FULL_PATH:
                path_to_create = container_target_path
                exit_code, _ = self.exec_cmd_in_container(f"ls {path_to_create}", fail_on_error=False)
            else:
                raise ValueError("Unknown create path mode: {}".format(create_container_path_mode))

            if exit_code != 0:
                self.create_dirs_in_container(path_to_create)

    def create_dirs_in_container(self, path_to_create):
        LOG.debug("Creating directories (recursive) '%s' in container '%s'", path_to_create, self.container.id)
        exit_code, _ = self.exec_cmd_in_container(f"mkdir -p {path_to_create}")
        if exit_code != 0:
            raise ValueError(
                "Failed to create directories '{}' in container {}".format(path_to_create, self.container.id)
            )


class DockerFileReplacer:
    vars_to_replace = {}

    # https://stackoverflow.com/a/30777398/1106893
    @classmethod
    def replace_all_vars(cls, input, vars_to_replace, default=None, skip_escaped=False):
        """Expand environment variables of form $var and ${var}.
        If parameter 'skip_escaped' is True, all escaped variable references
        (i.e. preceded by backslashes) are skipped.
        Unknown variables are set to 'default'. If 'default' is None,
        they are left unchanged.
        """
        cls.vars_to_replace = vars_to_replace

        def replace_var(m):
            # m.group(0) -> '${VAR}'
            # m.group(1) -> '{VAR}'
            # m.group(2) -> 'VAR'
            varname = m.group(2) or m.group(1)
            replaced_name = (
                DockerFileReplacer.vars_to_replace[varname] if varname in DockerFileReplacer.vars_to_replace else None
            )
            if not replaced_name:
                if default:
                    replaced_name = default
                else:
                    replaced_name = m.group(0)

            return replaced_name

        pattern = (r"(?<!\\)" if skip_escaped else "") + r"\$(\w+|\{([^}]*)\})"
        result = re.sub(pattern, replace_var, input)
        cls.vars_to_replace = {}

        # TODO print all remainder lines matching pattern -> Unresolved variables (LOG warning)
        # e.g. ${PYTHONPATH}
        return result

    @classmethod
    def add_env_var_declaration(cls, dockerfile_contents, var_name):
        lines = dockerfile_contents.split("\n")
        mod_lines = []
        for line in lines:
            if line.startswith("FROM"):
                mod_lines.append(line)
                mod_lines.append("ARG {var}".format(var=var_name))
                mod_lines.append('RUN echo "{var} = ${var}"'.format(var=var_name))
            else:
                mod_lines.append(line)

        return "\n".join(mod_lines)


class DockerCompose:
    COMPOSE_FILE_TEMPLATE = "docker-compose{profile}.yml"

    @staticmethod
    def up(working_dir, profile="", wait=0):
        if profile:
            profile = "-" + profile
        compose_file = DockerCompose.COMPOSE_FILE_TEMPLATE.format(profile=profile)
        command = "docker-compose -f {cfile} up -d".format(cfile=compose_file)
        SubprocessCommandRunner.run(
            command,
            working_dir=working_dir,
            log_command_result=True,
            fail_on_error=True,
            wait_after=10,
            wait_message="for docker compose command",
        )

    @staticmethod
    def logs(working_dir):
        command = "docker-compose logs"
        SubprocessCommandRunner.run(command, working_dir=working_dir, log_stdout=True)



import logging
import os
import time
from dataclasses import dataclass
from typing import List, Any, Dict

import requests
from jira import JIRA, JIRAError, Issue
from jira.resources import Attachment
from pythoncommons.git_utils import GitUtils

from pythoncommons.file_utils import FileUtils
from pythoncommons.github_utils import GithubPRMergeStatus
from pythoncommons.string_utils import StringUtils

LOG = logging.getLogger(__name__)


@dataclass
class JiraStatus:
    status: str
    resolution: str
    status_category: str


class JiraFetchMode:
    GSHEET = "GSHEET"
    ISSUES_CMDLINE = "ISSUES_CMDLINE"


class PatchOwner:
    def __init__(self, name, display_name):
        self.name = name
        self.display_name = display_name

    def __repr__(self):
        return repr((self.name, self.display_name))

    def __str__(self):
        # TODO understand unicode conversion issue in more details
        # return self.__class__.__name__ + \
        #        " { name: " + self.name + \
        #        ", display_name: " + str(self.display_name) + " }"
        # UnicodeEncodeError: 'ascii' codec can't encode character u'\xe1' in position 7: ordinal not in range(128)
        return (
            self.__class__.__name__
            + " { name: "
            + self.name
            + ", display_name: "
            + StringUtils.replace_special_chars(self.display_name)
            + " }"
        )


class JiraPatchStatus:
    APPLIES_CLEANLY = "APPLIES CLEANLY"
    CONFLICT = "CONFLICT"
    PATCH_ALREADY_COMMITTED = "PATCH_ALREADY_COMMITTED"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"
    PR_MERGEABLE = "PULL REQUEST MERGEABLE"
    PR_NOT_MERGEABLE = "PULL REQUEST NOT MERGEABLE"
    PR_STATUS_UNKNOWN = "PULL REQUEST UNKNOWN STATUS"
    CANNOT_FIND_PATCH_OR_PR = "PULL REQUEST OR PATCH NOT FOUND"
    ALLOWED_VALUES = {
        APPLIES_CLEANLY,
        CONFLICT,
        PATCH_ALREADY_COMMITTED,
        UNKNOWN_ERROR,
        PR_MERGEABLE,
        PR_NOT_MERGEABLE,
        PR_STATUS_UNKNOWN,
        CANNOT_FIND_PATCH_OR_PR,
    }

    @staticmethod
    def translate_from_github_pr_merge_statuses(
        github_merge_statuses: Dict[str, GithubPRMergeStatus]
    ) -> Dict[str, Any]:
        res = {}
        for branch, status in github_merge_statuses.items():
            res[branch] = JiraPatchStatus._translate_from_github_merge_status(status)
        return res

    @staticmethod
    def _translate_from_github_merge_status(github_merge_status):
        if github_merge_status == GithubPRMergeStatus.MERGEABLE:
            return JiraPatchStatus.PR_MERGEABLE
        elif github_merge_status == GithubPRMergeStatus.NOT_MERGEABLE:
            return JiraPatchStatus.PR_NOT_MERGEABLE
        elif github_merge_status == GithubPRMergeStatus.PR_NOT_FOUND:
            return JiraPatchStatus.CANNOT_FIND_PATCH_OR_PR
        elif github_merge_status == GithubPRMergeStatus.UNKNOWN:
            return JiraPatchStatus.PR_STATUS_UNKNOWN


class PatchApplicability:
    def __init__(self, applicable, reason=None, explicit=True):
        self.applicable = applicable
        self.explicit = explicit
        self.reason = reason
        if not applicable and not reason:
            raise ValueError("Reason should be specified is Patch is not applicable!")

    def __repr__(self):
        return repr((self.applicable, self.reason))

    def __str__(self):
        return self.__class__.__name__ + " { applicable: " + str(self.applicable) + ", reason: " + self.reason + " }"


class PatchApply:
    def __init__(self, patch, branch, result, conflicts=0, conflict_details=None):
        self.patch = patch
        self.branch = branch
        local_branch = GitUtils.convert_remote_branch_name_to_local(branch)
        if patch:
            self.explicit = patch.get_applicability(local_branch).explicit
        else:
            self.explicit = None

        if result not in JiraPatchStatus.ALLOWED_VALUES:
            raise ValueError("result must be a value found in PatchStatus!")

        if result != JiraPatchStatus.CONFLICT and conflicts > 0:
            raise ValueError(
                "Number of conflicts should be specified only if value of result is 'PatchStatus.CONFLICT'!"
            )
        if result != JiraPatchStatus.CONFLICT and conflict_details and len(conflict_details) > 0:
            raise ValueError("Conflict details should be specified only if value of result is 'PatchStatus.CONFLICT'!")

        self.result = result
        self.conflicts = conflicts
        self.conflict_details = conflict_details

    def __repr__(self):
        return repr((self.patch, self.branch, self.result, self.conflicts, self.conflict_details))

    def __str__(self):
        return (
            self.__class__.__name__
            + " { patch: "
            + self.patch
            + ", branch: "
            + str(self.branch)
            + ", result: "
            + str(self.result)
            + ", conflicts: "
            + str(self.conflicts)
            + " }"
        )


class PatchOverallStatus:
    def __init__(self, status):
        self.status = status

    def __repr__(self):
        return repr(self.status)

    def __str__(self):
        return self.__class__.__name__ + " { status: " + self.status + "}"


class JiraPatch:
    def __init__(self, issue_id, owner, patch_file):
        self.issue_id = issue_id
        # TODO owner and owner_short are currently not queried anywhere except __str__
        self.owner = owner
        self.owner_short = owner.name
        self.owner_display_name = owner.display_name
        self.filename = patch_file
        self.overall_status = PatchOverallStatus("N/A")
        self.file_path = None

    def set_patch_file_path(self, file_path):
        self.file_path = file_path

    def set_overall_status(self, overall_status):
        self.overall_status = overall_status

    def __repr__(self):
        return repr((self.issue_id, self.owner, self.filename))

    def __str__(self):
        return (
            self.__class__.__name__
            + " { issue_id: "
            + self.issue_id
            + ", owner: "
            + str(self.owner)
            + ", filename: "
            + str(self.filename)
            + " }"
        )

    def __hash__(self):
        return hash((self.issue_id, self.owner, self.filename))

    def __eq__(self, other):
        if isinstance(other, JiraPatch):
            return self.issue_id == other.issue_id and self.owner == other.owner and self.filename == other.filename
        return False


class AdvancedJiraPatch(JiraPatch):
    def __init__(self, issue_id, owner, version, target_branch, patch_file, applicability):
        super(AdvancedJiraPatch, self).__init__(issue_id, owner, patch_file)
        self.issue_id = issue_id
        # TODO owner and owner_short are currently not queried anywhere except __str__
        self.version = version
        self.target_branches = [target_branch]
        self.applicability = {target_branch: applicability}
        self.overall_status = PatchOverallStatus("N/A")

    def get_applicability(self, branch):
        return self.applicability[branch]

    def add_additional_branch(self, branch, applicability):
        self.target_branches.append(branch)
        self.applicability[branch] = applicability

    def is_applicable_for_branch(self, branch):
        if branch in self.applicability:
            return self.applicability[branch].applicable
        return False

    def get_reason_for_non_applicability(self, branch):
        if branch in self.applicability:
            return self.applicability[branch].reason
        return "Unknown"

    def is_applicable(self):
        applicabilities = set([True if a.applicable else False for a in self.applicability.values()])
        LOG.debug("Patch applicabilities: %s for patch %s", applicabilities, self)
        return True in applicabilities

    # TODO verify these
    def __repr__(self):
        return super(AdvancedJiraPatch, self).__repr__() + repr((self.version, self.target_branches))

    # TODO verify these
    def __str__(self):
        return super().__str__() + "version: " + str(self.version) + ", target_branch: " + str(self.target_branches)

    def __hash__(self):
        return hash((self.issue_id, self.owner, self.filename, tuple(self.target_branches)))

    def __eq__(self, other):
        if isinstance(other, AdvancedJiraPatch):
            return super().__eq__(other) and self.target_branches == other.target_branches
        return False


class JiraWrapper:
    def __init__(self, jira_url: str, default_branch: str, patches_root):
        options = {"server": jira_url}
        self.jira: JIRA = JIRA(options=options, timeout=20, max_retries=10)
        self.jira_url: str = jira_url
        self.default_branch: str = default_branch
        self.patches_root: str = patches_root

    def download_patch_file(self, patch: JiraPatch):
        LOG.debug("Querying jira issue %s", patch.issue_id)
        issue: Issue = self.jira.issue(patch.issue_id)

        found: bool = False
        for attachment in issue.fields.attachment:  # type: Attachment
            if patch.filename == attachment.filename:
                patch_file_path: str = os.path.join(self.patches_root, patch.issue_id, patch.filename)

                issue_dir: str = os.path.dirname(patch_file_path)
                if not os.path.exists(issue_dir):
                    os.makedirs(issue_dir)

                LOG.debug("Downloading patch from issue %s to file %s", patch.issue_id, patch_file_path)
                attachment_data = self.download_attachment_with_retries(attachment)

                # TODO let JiraPatch object create the path
                patch.set_patch_file_path(patch_file_path)
                FileUtils.write_to_file(patch_file_path, attachment_data, bytes=True)
                found = True
                break

        if not found:
            raise ValueError(
                "Cannot find attachment with name '{name}' for issue {issue}".format(
                    name=patch.filename, issue=patch.issue_id
                )
            )

    def get_jira_issue(self, issue_id: str):
        retries = 1
        try:
            if retries > 5:
                raise Exception("Jira could not be accessed 5 times, consecutively! Stopping execution.")
            return self.jira.issue(issue_id)
        except JIRAError:
            LOG.exception("JIRAError caught! Retrying to access jira!")
            retries += 1

    def download_attachment_with_retries(self, attachment: Attachment):
        tried = 0
        max_retries = 5
        while max_retries != tried:
            try:
                tried += 1
                return attachment.get()
            except requests.exceptions.Timeout:
                LOG.error("Read timed out while communicating with %s, sleeping for 5 seconds...", self.jira_url)
                time.sleep(5)

    def list_attachments(self, issue_id: str):
        issue = self.jira.issue(issue_id)

        for attachment in issue.fields.attachment:
            print("Name: '{filename}', size: {size}".format(filename=attachment.filename, size=attachment.size))
            # to read content use `get` method:
            # print("Content: '{}'".format(attachment.get()))

    def get_links(self, issue_id: str):
        issue: Issue = self.jira.issue(issue_id)
        return issue.fields.issuelinks

    def get_status(self, issue_id: str):
        issue = self.jira.issue(issue_id)
        status = issue.fields.status
        LOG.debug("Status of issue %s: %s", issue_id, status)
        return status.name

    def is_status_resolved(self, issue_id: str):
        status = self.get_status(issue_id)
        if status == "Resolved":
            LOG.debug("Status of jira is 'Resolved': %s", issue_id)
            return True
        return False

    @staticmethod
    def determine_patch_owner(jira_issue: Issue):
        if jira_issue.fields.assignee:
            owner_name = jira_issue.fields.assignee.name
            owner_display_name = jira_issue.fields.assignee.displayName
        else:
            owner_name = "unassigned"
            owner_display_name = "unassigned"
        return PatchOwner(owner_name, owner_display_name)

    def search(self, search_jql, max_results=999):
        return self.jira.search_issues(search_jql, maxResults=max_results)

    def get_subjiras_of_umbrella(self, jira_id):
        return self.search(f"parent = {jira_id}")

    def get_subjira_statuses_of_umbrella(self, jira_id) -> Dict[str, JiraStatus]:
        jiras: List[Any] = self.search(f"parent = {jira_id}")
        return {
            jira.key: JiraStatus(
                jira.fields.status.name, jira.fields.resolution, jira.fields.status.statusCategory.name
            )
            for jira in jiras
        }



import io
from dataclasses import dataclass
from typing import Callable, Any, List
from subprocess import Popen
import sh
import logging
import os
import pty
import select
import shlex
import subprocess
import sys
import termios
import time
import tty
from pythoncommons.file_utils import FileUtils
from pythoncommons.string_utils import auto_str, StringUtils


# TODO consolidate methods
class CommandRunner:
    def __init__(self, log):
        self.LOG = log

    def run_sync(
        self,
        cmd: str,
        fail_on_error: bool = False,
        single_result=True,
        _out: str = None,
        _err: str = None,
        _tee: bool = False,
    ):
        def _prepare_kwargs(_err, _out, _tee):
            kwargs = {}
            if _out:
                kwargs["_out"] = _out
            if _err:
                kwargs["_err"] = _err
            if _tee:
                kwargs["_tee"] = True
            return kwargs

        self.LOG.info("Running command: {}".format(cmd))
        output = []
        exit_code = None
        try:
            kwargs = _prepare_kwargs(_err, _out, _tee)
            if "_out" not in kwargs:
                # TODO There were problems passing this lambda from _prepare_kwargs
                process = sh.bash("-c", cmd, **kwargs, _out=lambda line: output.append(line))
            else:
                process = sh.bash("-c", cmd, **kwargs)
            process.wait()
            exit_code = process.exit_code
        except sh.ErrorReturnCode as e:
            if fail_on_error:
                raise e
            return e.stdout, e.exit_code
        except Exception as e:
            self.LOG.error("Error while executing command {}:\n {}".format(cmd, str(e)))

        self.LOG.info(" ".join(output))
        # Remove trailing newlines from each line
        output = [line.rstrip() for line in output]
        if len(output) == 1 and single_result:
            output = output[0]

        return output, exit_code

    def run_async(
        self, cmd: str, stdout_callback: Callable[[str], str] = None, stderr_callback: Callable[[str], str] = None
    ) -> Popen:
        self.LOG.info("Running command async: {}".format(cmd))

        _stdout_callback = stdout_callback
        _stderr_callback = stderr_callback

        if stdout_callback is None:
            _stdout_callback = print
        if stderr_callback is None:
            _stderr_callback = print

        process = sh.bash("-c", cmd, _out=_stdout_callback, _err=_stderr_callback, _bg=True)

        self.LOG.info("\n")

        return process

    def run_interactive(self, cmd: str) -> int:
        return subprocess.call(cmd, shell=True)

    @staticmethod
    def run(command, shell=False, shlex_split=True):
        if shlex_split:
            args = shlex.split(command)
        else:
            args = command
        proc = subprocess.run(
            args, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell
        )
        args2 = str(proc.args)
        return RegularCommandResult(command, args2, proc.stdout, proc.stderr, proc.returncode)

    @staticmethod
    def egrep_with_cli(
        git_log_result: List[str],
        file: str,
        grep_for: str,
        escape_single_quotes=True,
        escape_double_quotes=True,
        fail_on_empty_output=True,
        fail_on_error=True,
    ):
        FileUtils.save_to_file(file, StringUtils.list_to_multiline_string(git_log_result))
        if escape_single_quotes or escape_double_quotes:
            grep_for = StringUtils.escape_str(
                grep_for, escape_single_quotes=escape_single_quotes, escape_double_quotes=escape_double_quotes
            )
        cli_command = f'cat {file} | egrep "{grep_for}"'
        return CommandRunner.run_cli_command(
            cli_command, fail_on_empty_output=fail_on_empty_output, fail_on_error=fail_on_error
        )

    @staticmethod
    def execute_script(script: str, args: str, working_dir: str = None, output_file: str = None, use_tee=False):
        cli_command = ""
        if working_dir:
            cli_command += f"cd {working_dir};"
        cli_command += script
        if args:
            cli_command += " " + args
        if output_file:
            if use_tee:
                cli_command += f" | tee {output_file}"
            else:
                cli_command += f" > {output_file}"
        return CommandRunner.run_cli_command(cli_command)

    @staticmethod
    def run_cli_command(cli_command, fail_on_empty_output=True, print_command=True, fail_on_error=True):
        if print_command:
            LOG.info("Running CLI command: %s", cli_command)
        output = CommandRunner._getoutput(cli_command, raise_on_error=fail_on_error)
        if fail_on_empty_output and not output:
            # TODO Faulty format string: Search for all kinds of '%s' occurrences with ValueError
            raise ValueError("Command failed: %s", cli_command)
        return cli_command, output

    @staticmethod
    def _getoutput(command, raise_on_error=True):
        statusoutput = subprocess.getstatusoutput(command)
        if raise_on_error and statusoutput[0] != 0:
            raise ValueError(f"Command failed with exit code {statusoutput[0]}. Command was: {command}")
        return statusoutput[1]


LOG = logging.getLogger(__name__)
CONSOLE = logging.getLogger("console")


@dataclass
class RegularCommandResult:
    cli_cmd: str
    args: List[Any]
    stdout: str
    stderr: str
    exit_code: int


class SubprocessCommandRunner:
    @classmethod
    def run(
        cls,
        command,
        working_dir=None,
        log_stdout=False,
        log_stderr=False,
        log_command_result=False,
        fail_on_error=False,
        fail_message="",
        wait_after=0,
        wait_message="",
    ):
        if working_dir:
            FileUtils.change_cwd(working_dir)

        args = shlex.split(command)
        proc = subprocess.run(args, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        args2 = str(proc.args)

        if working_dir:
            FileUtils.reset_cwd()
        command_result = RegularCommandResult(command, args2, proc.stdout, proc.stderr, proc.returncode)

        if log_command_result:
            LOG.info("Command result: %s", command_result)
        if log_stdout:
            LOG.info("Stdout:\n %s", command_result.stdout)
        if log_stderr:
            LOG.info("Stderr:\n %s", command_result.stderr)

        if fail_on_error and command_result.exit_code != 0:
            if fail_message:
                LOG.error(fail_message)
            raise ValueError("Execution failed! Command was: {}".format(command))

        if wait_after > 0:
            LOG.info("Waiting for %d seconds %s", wait_message)
            time.sleep(wait_after)

    @classmethod
    def run_and_follow_stdout_stderr(
        cls, cmd, log_file=FileUtils.get_temp_file_name(), stdout_logger=None, exit_on_nonzero_exitcode=False
    ):
        # TODO stderr is not logged at all
        if not stdout_logger:
            stdout_logger = LOG
        args = shlex.split(cmd)
        LOG.info(f"Running command: {cmd}")
        LOG.info(f"Command args: {args}")
        LOG.info(f"Config: Logging stderr to stdout, and stdout to logger. The logger is: {stdout_logger}")
        LOG.info(f"Config: Also logging to file: {log_file}")

        with open(log_file, "w") as f:
            # Redirect stderr to stdout
            process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            LOG.info(f"Returned process from subprocess.Popen: {process}")
            stdout_wrapper = io.TextIOWrapper(process.stdout, encoding="utf-8")
            for line in stdout_wrapper:
                line = line.strip()
                stdout_logger.info(line)
                f.write(line + os.linesep)
            stdout_wrapper.close()
            while process.poll() is None:
                LOG.info("Waiting for process to terminate...")
                time.sleep(2)
            LOG.info(f"Exit code of command '{cmd}' was: {process.returncode}")

            if exit_on_nonzero_exitcode and process.returncode != 0:
                LOG.error(f"Non-zero exit code was received while running command: {cmd}. Please check logs above.")
                sys.exit(process.returncode)
            return process


# TODO Warning: untested
class InteractiveCommandRunner:
    # https://intellij-support.jetbrains.com/hc/en-us/community/posts/360003383619-Pycharm-2019-termios-error-25-Inappropriate-ioctl-for-device-
    # In PyCharm, in Run configurations, this must be enabled: Emulate terminal in output console
    def run(self, command):
        if not command:
            raise ValueError("Command was empty or None!")

        LOG.info("Executing Command: " + str(command))
        CONSOLE.info("Command: " + str(command))
        # command = 'docker run -it --rm centos /bin/bash'.split()

        # save original tty setting then set it to raw mode
        old_tty = termios.tcgetattr(sys.stdin)
        tty.setraw(sys.stdin.fileno())

        # open pseudo-terminal to interact with subprocess
        master_fd, slave_fd = pty.openpty()

        # use os.setsid() make it run in a new process group, or bash job control will not be enabled

        p = subprocess.Popen(
            command,
            preexec_fn=os.setsid,
            stdin=slave_fd,
            stdout=slave_fd,
            stderr=slave_fd,
            universal_newlines=True,
            shell=True,
        )

        while p.poll() is None:
            r, w, e = select.select([sys.stdin, master_fd], [], [])
            if sys.stdin in r:
                d = os.read(sys.stdin.fileno(), 10240)
                os.write(master_fd, d)
            elif master_fd in r:
                o = os.read(master_fd, 10240)
                if o:
                    os.write(sys.stdout.fileno(), o)

        # restore tty settings back
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)



from typing import List, Tuple

from bs4 import BeautifulSoup


class HtmlGenerator:
    def __init__(self):
        self.last_html_separator = None
        self.html = None
        self.finished_html = False
        self.soup = BeautifulSoup()

    def begin_html_tag(self):
        self.html = self.soup.new_tag("html")
        return self

    def finish_html(self):
        if self.finished_html:
            raise ValueError("Invalid state. Already finished html.")
        if not self.html:
            raise ValueError("Invalid state. Please call 'begin_html_tag' first.")

        self.soup.append(self.html)
        self.finished_html = True
        return self

    def render(self):
        if not self.finished_html:
            self.finish_html()
        return self.soup.prettify()

    def add_basic_table_style(self):
        if not self.html:
            raise ValueError("Invalid state. Please call 'begin_html_tag' first.")

        head = self.soup.new_tag("head")
        style = self.soup.new_tag("style")
        style.string = """
table, th, td {
  border: 1px solid black;
}
"""
        head.append(style)
        self.html.append(head)
        return self

    def append_paragraphs(self, strings: List[str]):
        for line in strings:
            self.append_paragraph(line)
        return self

    def append_paragraph(self, string):
        p = self.soup.new_tag("p")
        p.append(string)
        self.soup.append(p)
        return self

    @staticmethod
    def generate_separator(tag="hr", breaks=2):
        html_sep: str = f"<{tag}/>"
        html_sep += breaks * "<br/>"
        return html_sep

    def append_html_tables(self,
                           table_tups: List[Tuple[str, str]],
                           separator=None,
                           additional_separator_at_beginning=False,
                           header_type="h1"):
        if not self.html:
            raise ValueError("Invalid state. Please call 'begin_html_tag' first.")

        tables_html = ""
        if not separator:
            separator = ""
        ht = header_type

        if additional_separator_at_beginning:
            tables_html += separator

        # header = tup[0], table = tup[1]
        tables_html += separator.join([f"<h1>{tup[0]}</h1>" + tup[1] for tup in table_tups])
        gen_tables_soup = BeautifulSoup(tables_html, "html.parser")
        self.html.append(gen_tables_soup)
        return self
